from typing import List, Tuple, Union, Set, Optional
import numpy as np
import torch
from torch import nn, Tensor
import torch.nn.functional as F
from deepxube.utils import misc_utils
import re
from random import randrange
import matplotlib.pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
import matplotlib.patches as patches

from deepxube.nnet.pytorch_models import ResnetModel, FullyConnectedModel
from .environment_abstract import EnvGrndAtoms, State, Action, Goal, HeurFnNNet
from deepxube.logic.logic_objects import Atom, Model
from numpy.typing import NDArray


int_t = Union[np.uint8, np.int_]


class NPState(State):
    __slots__ = ['tiles', 'hash']

    def __init__(self, tiles: NDArray[int_t]):
        self.tiles: NDArray[int_t] = tiles
        self.hash: Optional[int] = None

    def __hash__(self):
        if self.hash is None:
            self.hash = hash(self.tiles.tobytes())
        return self.hash

    def __eq__(self, other: object):
        if isinstance(other, NPState):
            return np.array_equal(self.tiles, other.tiles)
        return NotImplemented


class NPGoal(Goal):
    def __init__(self, tiles: NDArray[int_t]):
        self.tiles: NDArray[int_t] = tiles


class NPAction(Action):
    def __init__(self, action: int):
        self.action = action

    def __hash__(self):
        return self.action

    def __eq__(self, other: object):
        if isinstance(other, NPAction):
            return self.action == other.action
        return NotImplemented


class ProcessStates(nn.Module):
    def __init__(self, state_dim: int, one_hot_depth: int):
        super().__init__()
        self.state_dim: int = state_dim
        self.one_hot_depth: int = one_hot_depth

    def forward(self, states_nnet: Tensor):
        x = states_nnet

        # preprocess input
        if self.one_hot_depth > 0:
            x = F.one_hot(x.long(), self.one_hot_depth)
            x = x.float()
            x = x.view(-1, self.state_dim * self.one_hot_depth)
        else:
            x = x.float()

        return x


class FCResnet(nn.Module):
    def __init__(self, input_dim: int, h1_dim: int, resnet_dim: int, num_resnet_blocks: int, out_dim: int,
                 batch_norm: bool, weight_norm: bool):
        super().__init__()
        self.first_fc = FullyConnectedModel(input_dim, [h1_dim, resnet_dim], [batch_norm] * 2, ["RELU"] * 2,
                                            weight_norms=[weight_norm] * 2)
        self.resnet = ResnetModel(resnet_dim, num_resnet_blocks, out_dim, batch_norm, weight_norm=weight_norm,
                                  layer_act="RELU")

    def forward(self, x: Tensor):
        x = self.first_fc(x)
        x = self.resnet(x)

        return x


class NNet(HeurFnNNet):
    def __init__(self, state_dim: int, one_hot_depth: int, h1_dim: int, resnet_dim: int, num_res_blocks: int,
                 out_dim: int, batch_norm: bool, weight_norm: bool, nnet_type: str):
        super().__init__(nnet_type)
        self.state_proc = ProcessStates(state_dim, one_hot_depth)

        input_dim: int = state_dim * one_hot_depth * 2
        self.heur = FCResnet(input_dim, h1_dim, resnet_dim, num_res_blocks, out_dim, batch_norm, weight_norm)

    def forward(self, states_goals_l: List[Tensor]):
        states_proc = self.state_proc(states_goals_l[0])
        goals_proc = self.state_proc(states_goals_l[1])

        x: Tensor = self.heur(torch.cat((states_proc, goals_proc), dim=1))

        return x


class NPuzzle(EnvGrndAtoms[NPState, NPAction, NPGoal]):
    moves: List[str] = ['U', 'D', 'L', 'R']
    moves_rev: List[str] = ['D', 'U', 'R', 'L']

    def __init__(self, env_name: str, dim: int):
        super().__init__(env_name)

        self.dim: int = dim
        self.dtype: type
        if self.dim <= 15:
            self.dtype = np.uint8
        else:
            self.dtype = np.int_

        self.num_tiles: int = dim ** 2

        # Solved state
        self.goal_tiles: NDArray[int_t] = np.concatenate((np.arange(1, self.dim * self.dim),
                                                          [0])).astype(self.dtype)  # type: ignore

        # Next state ops
        self.swap_zero_idxs: NDArray[int_t] = self._get_swap_zero_idxs(self.dim)

        self.num_actions: int = 4

    def next_state(self, states: List[NPState], actions: List[NPAction]) -> Tuple[List[NPState], List[float]]:
        # initialize
        states_np: NDArray[int_t] = np.stack([x.tiles for x in states], axis=0)
        states_next_np: NDArray[int_t] = states_np.copy()

        # get zero indicies
        z_idxs: NDArray[np.int_]
        _, z_idxs = np.where(states_next_np == 0)

        tcs_np: NDArray[np.float64] = np.zeros(len(states))
        for action in set(actions):
            action_idxs: NDArray[np.int_] = np.array([idx for idx in range(len(actions)) if actions[idx] == action])
            states_np_act = states_np[action_idxs]
            z_idxs_act: NDArray[np.int_] = z_idxs[action_idxs]

            states_next_np_act, _, tcs_act = self._move_np(states_np_act, z_idxs_act, action.action)

            states_next_np[action_idxs] = states_next_np_act
            tcs_np[action_idxs] = np.array(tcs_act)

        # make states
        states_next: List[NPState] = [NPState(x) for x in list(states_next_np)]
        transition_costs = list(tcs_np)

        return states_next, transition_costs

    def expand(self, states: List[NPState]) -> Tuple[List[List[NPState]], List[List[NPAction]], List[List[float]]]:
        # initialize
        num_states: int = len(states)

        states_exp: List[List[NPState]] = [[] for _ in range(len(states))]
        actions_exp_l: List[List[NPAction]] = [[] for _ in range(len(states))]

        tc: NDArray[np.float64] = np.empty([num_states, self.num_actions])

        # numpy states
        states_np: NDArray[int_t] = np.stack([state.tiles for state in states])

        # Get z_idxs
        z_idxs: NDArray[np.int_]
        _, z_idxs = np.where(states_np == 0)

        # for each move, get next states, transition costs, and if solved
        for action in range(self.num_actions):
            # next state
            states_next_np: NDArray[int_t]
            tc_move: List[float]
            states_next_np, _, tc_move = self._move_np(states_np, z_idxs, action)

            # transition cost
            tc[:, action] = np.array(tc_move)

            for idx in range(len(states)):
                states_exp[idx].append(NPState(states_next_np[idx]))
                actions_exp_l[idx].append(NPAction(action))

        # make lists
        tc_l: List[List[float]] = [list(tc[i]) for i in range(num_states)]

        return states_exp, actions_exp_l, tc_l

    def get_state_actions(self, states: List[NPState]) -> List[List[NPAction]]:
        return [[NPAction(x) for x in range(self.num_actions)] for _ in range(len(states))]

    def is_solved(self, states: List[NPState], goals: List[NPGoal]) -> List[bool]:
        states_np = np.stack([x.tiles for x in states], axis=0)
        goals_np = np.stack([x.tiles for x in goals], axis=0)
        is_solved_np = np.all(np.logical_or(states_np == goals_np, goals_np == self.num_tiles), axis=1)
        return list(is_solved_np)

    def states_goals_to_nnet_input(self, states: List[NPState], goals: List[NPGoal]) -> List[NDArray[int_t]]:
        states_np: NDArray[int_t] = np.stack([x.tiles for x in states], axis=0)
        goals_np: NDArray[int_t] = np.stack([x.tiles for x in goals], axis=0)

        return [states_np.astype(self.dtype), goals_np]

    def state_to_model(self, states: List[NPState]) -> List[Model]:
        states_np: NDArray[int_t] = np.stack([state.tiles for state in states], axis=0).astype(self.dtype)
        states_np = states_np.reshape((-1, self.dim, self.dim))
        models: List[Model] = [self._sqr_tiles_to_model(x) for x in states_np]
        return models

    def model_to_state(self, states_m: List[Model]) -> List[NPState]:
        for state_m in states_m:
            assert len(state_m) == self.num_tiles, "model should be fully specified"
        return [NPState(x) for x in self._models_to_np(states_m)]

    def goal_to_model(self, goals: List[NPGoal]) -> List[Model]:
        goals_np: NDArray[int_t] = np.stack([goal.tiles for goal in goals], axis=0).astype(self.dtype)
        goals_np = goals_np.reshape((-1, self.dim, self.dim))
        models: List[Model] = [self._sqr_tiles_to_model(x) for x in goals_np]
        return models

    def model_to_goal(self, models: List[Model]) -> List[NPGoal]:
        return [NPGoal(x) for x in self._models_to_np(models)]

    def get_v_nnet(self) -> HeurFnNNet:
        nnet = NNet(self.num_tiles, self.num_tiles + 1, 5000, 1000, 4, 1, True, False, "V")
        return nnet

    def get_q_nnet(self) -> HeurFnNNet:
        nnet = NNet(self.num_tiles, self.num_tiles + 1, 5000, 1000, 4, self.num_actions, True, False, "V")
        return nnet

    def get_start_states(self, num_states: int) -> List[NPState]:
        assert (num_states > 0)
        states: List[NPState] = []
        while len(states) < num_states:
            states_np: NDArray[int_t] = np.stack([np.random.permutation(self.num_tiles)
                                                  for _ in range(num_states - len(states))], axis=0).astype(self.dtype)
            is_solvable: NDArray[np.bool_] = self._is_solvable(states_np)

            states.extend([NPState(x) for x in states_np[is_solvable]])

        return states

    def start_state_fixed(self, states: List[NPState]) -> List[Model]:
        return [frozenset() for _ in states]

    def get_pddl_domain(self) -> List[str]:
        pddl_str: str = """
        (define (domain strips-sliding-tile)
  (:requirements :strips)
  (:predicates
   (tile ?x) (position ?x)
   (at ?t ?x ?y) (blank ?x ?y)
   (inc ?p ?pp) (dec ?p ?pp)
   (up ?t) (down ?t) (left ?t) (right ?t)
   )

  (:action move-up
    :parameters (?omf ?px ?py ?by)
    :precondition (and (up ?omf)
        (tile ?omf) (position ?px) (position ?py) (position ?by)
        (dec ?by ?py) (blank ?px ?by) (at ?omf ?px ?py))
    :effect (and (not (blank ?px ?by)) (not (at ?omf ?px ?py))
    (blank ?px ?py) (at ?omf ?px ?by)))

  (:action move-down
    :parameters (?omf ?px ?py ?by)
    :precondition (and (down ?omf)
        (tile ?omf) (position ?px) (position ?py) (position ?by)
        (inc ?by ?py) (blank ?px ?by) (at ?omf ?px ?py))
    :effect (and (not (blank ?px ?by)) (not (at ?omf ?px ?py))
    (blank ?px ?py) (at ?omf ?px ?by)))

  (:action move-left
    :parameters (?omf ?px ?py ?bx)
    :precondition (and (left ?omf)
        (tile ?omf) (position ?px) (position ?py) (position ?bx)
        (dec ?bx ?px) (blank ?bx ?py) (at ?omf ?px ?py))
    :effect (and (not (blank ?bx ?py)) (not (at ?omf ?px ?py))
        (blank ?px ?py) (at ?omf ?bx ?py)))

  (:action move-right
    :parameters (?omf ?px ?py ?bx)
    :precondition (and (right ?omf)
        (tile ?omf) (position ?px) (position ?py) (position ?bx)
        (inc ?bx ?px) (blank ?bx ?py) (at ?omf ?px ?py))
    :effect (and (not (blank ?bx ?py)) (not (at ?omf ?px ?py))
    (blank ?px ?py) (at ?omf ?bx ?py)))
  ) """

        return pddl_str.split("\n")

    def state_goal_to_pddl_inst(self, state: NPState, goal: NPGoal) -> List[str]:
        model: Model = self.goal_to_model([goal])[0]

        # objects
        inst_l: List[str] = ["(define(problem slidingtile)", "(:domain strips-sliding-tile)"]
        tile_names = [f"t{i}" for i in range(1, self.num_tiles)]
        positions = [f"x{i + 1}" for i in range(0, self.dim)] + [f"y{i + 1}" for i in range(0, self.dim)]
        objects: List[str] = tile_names.copy() + positions.copy()

        inst_l.append(f"(:objects {' '.join(objects)})")

        # tiles and positions
        inst_l.append("(:init")
        tile_grnd_atoms: List[str] = [f"(tile {x})" for x in tile_names]
        position_grnd_atoms: List[str] = [f"(position {x})" for x in positions]
        inst_l.append(f"{' '.join(tile_grnd_atoms)}")
        inst_l.append(f"{' '.join(position_grnd_atoms)}")

        # inc and dec
        inc_grnd_atoms: List[str] = []
        for idx in range(self.dim - 1):
            inc_grnd_atoms.append(f"(inc x{idx + 1} x{idx + 2})")
            inc_grnd_atoms.append(f"(inc y{idx + 1} y{idx + 2})")

        dec_grnd_atoms: List[str] = []
        for idx in range(self.dim - 1):
            dec_grnd_atoms.append(f"(dec x{idx + 2} x{idx + 1})")
            dec_grnd_atoms.append(f"(dec y{idx + 2} y{idx + 1})")

        inst_l.append(f"{' '.join(inc_grnd_atoms)}")
        inst_l.append(f"{' '.join(dec_grnd_atoms)}")

        # state
        inst_l.append("")
        tiles_mat = state.tiles.reshape((self.dim, self.dim))
        for idx_y in range(tiles_mat.shape[0]):
            state_lits_row: List[str] = []
            for idx_x in range(tiles_mat.shape[1]):
                tile = tiles_mat[idx_y, idx_x]
                if tile == 0:
                    state_lits_row.append(f"(blank x{idx_x + 1} y{idx_y + 1})")
                else:
                    state_lits_row.append(f"(at t{tile} x{idx_x + 1} y{idx_y + 1})")
            inst_l.append(f"{' '.join(state_lits_row)}")
        inst_l.append("")

        # up, down, left, right
        for direction_name in ["up", "down", "left", "right"]:
            direction_pred_names: List[str] = [f"({direction_name} {x})" for x in tile_names]
            inst_l.append(f"{' '.join(direction_pred_names)}")
        inst_l.append(")")

        # goal
        inst_l.append("(:goal")
        if len(model) > 0:
            inst_l.append("(and")
            tiles_goal_mat = self._models_to_np([model])[0].reshape((self.dim, self.dim))
            for idx_y in range(tiles_goal_mat.shape[0]):
                goal_lits_row: List[str] = []
                for idx_x in range(tiles_goal_mat.shape[1]):
                    tile = tiles_goal_mat[idx_y, idx_x]
                    if tile == self.num_tiles:
                        goal_lits_row.append("                                     ")
                    elif tile == 0:
                        goal_lits_row.append(f"(blank x{idx_x + 1} y{idx_y + 1})")
                    else:
                        goal_lits_row.append(f"(at t{tile} x{idx_x + 1} y{idx_y + 1})")
                inst_l.append(f"{' '.join(goal_lits_row)}")

            inst_l.append(")")
        else:
            inst_l.append("(tile t1)")  # TODO hacky, how to do empty goal in PDDL?

        inst_l.append(")")
        inst_l.append(")")

        return inst_l

    def pddl_action_to_action(self, pddl_action: str) -> int:
        if re.match("^move-up", pddl_action):
            return 1
        elif re.match("^move-down", pddl_action):
            return 0
        elif re.match("^move-left", pddl_action):
            return 3
        elif re.match("^move-right", pddl_action):
            return 2

        raise ValueError(f"Unknown action {pddl_action}")

    def visualize(self, states: Union[List[NPState], List[NPGoal]]) -> NDArray[np.float64]:
        fig = plt.figure(figsize=(.64, .64))
        ax = fig.add_axes((0, 0, 1., 1.))
        # fig = plt.figure(figsize=(.64, .64))
        # ax = fig.gca()
        # fig.add_axes(ax)
        canvas = FigureCanvas(fig)
        width, height = fig.get_size_inches() * fig.get_dpi()
        width = int(width)
        height = int(height)
        states_img: NDArray[np.float64] = np.zeros((len(states), width, height, 3))
        for state_idx, state in enumerate(states):
            ax.clear()

            state_np: NDArray[int_t]
            if isinstance(state, NPState):
                state_np = state.tiles
            elif isinstance(state, NPGoal):
                model: Model = self.goal_to_model([state])[0]
                state_np = self._models_to_np([model])[0]
            else:
                raise ValueError(f"Unknown input type {type(state)}")

            for square_idx, square in enumerate(state_np):
                color = 'white'
                x_pos = int(np.floor(square_idx / self.dim))
                yPos = square_idx % self.dim

                left = yPos / float(self.dim)
                right = left + 1.0 / float(self.dim)
                top = (self.dim - x_pos - 1) / float(self.dim)
                bottom = top + 1.0 / float(self.dim)

                ax.add_patch(patches.Rectangle((left, top), 1.0 / self.dim, 1.0 / self.dim, linewidth=1,
                                               edgecolor='k', facecolor=color))

                if square != 0:
                    sqr_txt: str
                    if square == (self.dim ** 2):
                        sqr_txt = "-"
                    else:
                        sqr_txt = str(square)
                    ax.text(0.5 * (left + right), 0.5 * (bottom + top), sqr_txt, horizontalalignment='center',
                            verticalalignment='center', fontsize=6, color='black', transform=ax.transAxes)

            canvas.draw()
            states_img[state_idx] = np.frombuffer(canvas.tostring_rgb(),
                                                  dtype='uint8').reshape((width, height, 3)) / 255

        plt.close(fig)
        return states_img

    def get_ground_atoms(self) -> List[Atom]:
        ground_atoms: List[Atom] = []
        for tile_num in range(self.num_tiles):
            for idx_x in range(self.dim):
                for idx_y in range(self.dim):
                    ground_atoms.append(("at_idx", f"t{tile_num}", f"r{idx_x}", f"c{idx_y}"))

        return ground_atoms

    def on_model(self, m) -> Model:
        symbs_set: Set[str] = set(str(x) for x in m.symbols(shown=True))
        symbs: List[str] = [misc_utils.remove_all_whitespace(symb) for symb in symbs_set]

        # get atoms
        atoms: List[Atom] = []
        for symb in symbs:
            match = re.search(r"^at_idx\((\S+),(\S+),(\S+)\)$", symb)
            if match is None:
                continue
            atom: Atom = ("at_idx", match.group(1), match.group(2), match.group(3))
            atoms.append(atom)

        model: Model = frozenset(atoms)
        return model

    def get_bk(self) -> List[str]:
        bk: List[str] = ["%tiles and blanks"]

        for tile_num in range(1, self.num_tiles):
            bk.append(f"tile(t{tile_num})")
        bk.append("blank(t0)")
        bk.append("t_or_b(X) :- tile(X)")
        bk.append("t_or_b(X) :- blank(X)")

        for tile_num in range(0, self.num_tiles):
            bk.append(f"val(t{tile_num}, {tile_num})")
        bk.append(f"num(0..{sum(range(self.num_tiles + 1))})")

        bk.append("")
        bk.append("%rows and columns")
        for idx in range(self.dim):
            bk.append(f"row(r{idx})")
        for idx in range(self.dim):
            bk.append(f"col(c{idx})")
        bk.append("at_row(X, R) :- t_or_b(X), row(R), at_idx(X, R, _)")
        bk.append("at_col(X, C) :- t_or_b(X), col(C), at_idx(X, _, C)")
        bk.append(f"num_row({self.dim})")
        bk.append(f"num_col({self.dim})")

        bk.append("")
        bk.append("% classical negation")
        bk.append("-at_idx(X, R, C) :- t_or_b(X), t_or_b(X2), at_idx(X2, R, C), not X=X2")
        bk.append("-at_idx(X, R, C) :- row(R), col(C), row(R2), at_idx(X, R2, _), not R=R2")
        bk.append("-at_idx(X, R, C) :- row(R), col(C), col(C2), at_idx(X, _, C2), not C=C2")

        bk.append("")
        bk.append("% constraints")
        bk.append("% location cannot have multiple tiles")
        bk.append(":- row(R), col(C), #count{X: at_idx(X, R, C)} > 1")

        bk.append("% tile or blank cannot be in more than one place at a time")
        bk.append(":- t_or_b(X), #count{R, C: at_idx(X, R, C)} > 1")

        return bk

    def _is_solvable(self, states_np: NDArray[int_t]) -> NDArray[np.bool_]:
        num_inversions: NDArray[np.int_] = self._get_num_inversions(states_np)
        num_inversions_is_even: NDArray[np.bool_] = (num_inversions % 2 == 0)
        if self.dim % 2 == 0:
            # even
            _, z_idxs = np.where(states_np == 0)
            z_row_from_bottom_1 = self.dim - np.floor(z_idxs / self.dim)
            z_from_bottom_1_is_even: NDArray[np.bool_] = (z_row_from_bottom_1 % 2 == 0)
            case_1: NDArray[np.bool_] = np.logical_and(z_from_bottom_1_is_even, np.logical_not(num_inversions_is_even))
            case_2: NDArray[np.bool_] = np.logical_and(np.logical_not(z_from_bottom_1_is_even), num_inversions_is_even)
            return np.logical_or(case_1, case_2)
        else:
            # odd
            return num_inversions_is_even

    def _get_num_inversions(self, states_np: NDArray[int_t]) -> NDArray[np.int_]:
        num_inversions: NDArray[np.int_] = np.zeros(states_np.shape[0], dtype=int)
        for idx_1 in range(self.num_tiles):
            for idx_2 in range(idx_1 + 1, self.num_tiles):
                no_zeros: NDArray[np.bool_] = np.logical_and(states_np[:, idx_1] != 0, states_np[:, idx_2] != 0)
                has_inversion: NDArray[np.bool_] = states_np[:, idx_1] > states_np[:, idx_2]
                num_inversions = num_inversions + np.logical_and(no_zeros, has_inversion)

        return num_inversions

    def _sqr_tiles_to_model(self, tiles_sqr: NDArray[int_t]):
        grnd_atoms: List[Atom] = []
        for idx_x in range(tiles_sqr.shape[0]):
            for idx_y in range(tiles_sqr.shape[1]):
                val = tiles_sqr[idx_x, idx_y]
                if val != self.num_tiles:
                    grnd_atoms.append(('at_idx', f"t{tiles_sqr[idx_x, idx_y]}", f"r{str(idx_x)}", f"c{str(idx_y)}"))

        return frozenset(grnd_atoms)

    def _models_to_np(self, models: List[Model]) -> NDArray[int_t]:
        models_np: NDArray[int_t] = (np.ones((len(models), self.dim, self.dim)) * self.num_tiles).astype(self.dtype)
        for idx, model in enumerate(models):
            for grnd_atom in model:
                models_np[idx, int(grnd_atom[2][1:]), int(grnd_atom[3][1:])] = int(grnd_atom[1][1:])

        return models_np.reshape((len(models), -1))

    def _random_walk(self, states: List[NPState], num_steps_l: List[int]) -> List[NPState]:
        states_np = np.stack([x.tiles for x in states], axis=0)

        # Get z_idxs
        z_idxs: NDArray[np.int_]
        _, z_idxs = np.where(states_np == 0)

        # Scrambles
        num_steps_np: NDArray[np.int_] = np.array(num_steps_l)
        num_actions: NDArray[np.int_] = np.zeros(len(states), dtype=int)

        # go backward from goal state
        while int(np.max(num_actions < num_steps_np)) > 0:
            idxs: NDArray[np.int_] = np.where((num_actions < num_steps_np))[0]
            subset_size: int = int(max(len(idxs) / self.num_actions, 1))
            idxs = np.random.choice(idxs, subset_size)

            move: int = randrange(self.num_actions)
            states_np[idxs], z_idxs[idxs], _ = self._move_np(states_np[idxs], z_idxs[idxs], move)

            num_actions[idxs] = num_actions[idxs] + 1

        return [NPState(x) for x in states_np]

    def _get_swap_zero_idxs(self, n: int) -> NDArray[int_t]:
        swap_zero_idxs: NDArray[int_t] = np.zeros((n ** 2, len(NPuzzle.moves)), dtype=self.dtype)
        for moveIdx, move in enumerate(NPuzzle.moves):
            for i in range(n):
                for j in range(n):
                    z_idx = np.ravel_multi_index((i, j), (n, n))

                    state = np.ones((n, n), dtype=int)
                    state[i, j] = 0

                    is_eligible: bool = False
                    if move == 'U':
                        is_eligible = i < (n - 1)
                    elif move == 'D':
                        is_eligible = i > 0
                    elif move == 'L':
                        is_eligible = j < (n - 1)
                    elif move == 'R':
                        is_eligible = j > 0

                    if is_eligible:
                        swap_i: int = -1
                        swap_j: int = -1
                        if move == 'U':
                            swap_i = i + 1
                            swap_j = j
                        elif move == 'D':
                            swap_i = i - 1
                            swap_j = j
                        elif move == 'L':
                            swap_i = i
                            swap_j = j + 1
                        elif move == 'R':
                            swap_i = i
                            swap_j = j - 1

                        swap_zero_idxs[z_idx, moveIdx] = np.ravel_multi_index((swap_i, swap_j), (n, n))
                    else:
                        swap_zero_idxs[z_idx, moveIdx] = z_idx

        return swap_zero_idxs

    def _move_np(self, states_np: NDArray[int_t], z_idxs: NDArray[np.int_],
                 action: int) -> Tuple[NDArray[int_t], NDArray[int_t], List[float]]:
        states_next_np: NDArray[int_t] = states_np.copy()

        # get index to swap with zero
        state_idxs: NDArray[np.int_] = np.arange(0, states_next_np.shape[0]).astype(int)
        swap_z_idxs: NDArray[int_t] = self.swap_zero_idxs[z_idxs, action]

        # swap zero with adjacent tile
        states_next_np[state_idxs, z_idxs] = states_np[state_idxs, swap_z_idxs]
        states_next_np[state_idxs, swap_z_idxs] = 0

        # transition costs
        transition_costs: List[float] = [1.0 for _ in range(states_np.shape[0])]

        return states_next_np, swap_z_idxs, transition_costs



from typing import List, Tuple, Dict, Callable, Optional, cast
from deepxube.environments.environment_abstract import Environment, State, Action, Goal
from deepxube.nnet.nnet_utils import HeurFN_T
import numpy as np
from heapq import heappush, heappop

from deepxube.utils import misc_utils
import time
from numpy.typing import NDArray


class Node:
    __slots__ = ['state', 'goal', 'path_cost', 'heuristic', 'cost', 'is_solved', 'parent_action', 'parent']

    def __init__(self, state: State, goal: Goal, path_cost: float, is_solved: bool,
                 parent_action: Optional[Action], parent):
        self.state: State = state
        self.goal: Goal = goal
        self.path_cost: float = path_cost
        self.heuristic: Optional[float] = None
        self.cost: Optional[float] = None
        self.is_solved: bool = is_solved
        self.parent_action: Optional[Action] = parent_action
        self.parent: Optional[Node] = parent


OpenSetElem = Tuple[float, int, Node]


class Instance:

    def __init__(self, root_node: Node, weight: float):
        self.open_set: List[OpenSetElem] = []
        self.heappush_count: int = 0
        self.closed_dict: Dict[State, float] = dict()
        self.popped_nodes: List[Node] = []
        self.goal_node: Optional[Node] = None
        self.finished: bool = False
        self.weight: float = weight
        self.num_nodes_generated: int = 0
        self.step_num: int = 0

        self.root_node: Node = root_node

        self.push_to_open([self.root_node])

    def push_to_open(self, nodes: List[Node]):
        for node in nodes:
            heappush(self.open_set, (cast(float, node.cost), self.heappush_count, node))
            self.heappush_count += 1

    def pop_from_open(self, num_nodes: int) -> List[Node]:
        num_to_pop: int = min(num_nodes, len(self.open_set))

        popped_nodes = [heappop(self.open_set)[2] for _ in range(num_to_pop)]

        for node in popped_nodes:
            if node.is_solved and ((self.goal_node is None) or
                                   (cast(float, node.cost) < cast(float, self.goal_node.cost))):
                self.goal_node = node

        if (self.goal_node is not None) and (cast(float, self.goal_node.cost) <= cast(float, popped_nodes[0].cost)):
            self.finished = True
        self.popped_nodes.extend(popped_nodes)

        return popped_nodes

    def remove_in_closed(self, nodes: List[Node]) -> List[Node]:
        nodes_not_in_closed: List[Node] = []

        for node in nodes:
            path_cost_prev: Optional[float] = self.closed_dict.get(node.state)
            if path_cost_prev is None:
                nodes_not_in_closed.append(node)
                self.closed_dict[node.state] = node.path_cost
            elif path_cost_prev > node.path_cost:
                nodes_not_in_closed.append(node)
                self.closed_dict[node.state] = node.path_cost

        return nodes_not_in_closed


def expand_nodes(instances: List[Instance], popped_nodes_all: List[List[Node]], env: Environment):
    # Get children of all nodes at once (for speed)
    popped_nodes_flat: List[Node]
    split_idxs: List[int]
    popped_nodes_flat, split_idxs = misc_utils.flatten(popped_nodes_all)

    if len(popped_nodes_flat) == 0:
        return [[]]

    states: List[State] = [x.state for x in popped_nodes_flat]

    states_c_by_node: List[List[State]]
    tcs_np: List[List[float]]
    states_c_by_node, actions_c_by_node, tcs_np = env.expand(states)

    tcs_by_node: List[List[float]] = [list(x) for x in tcs_np]
    goals_c_by_node: List[List[Goal]] = [[node.goal] * len(states_c) for node, states_c in
                                         zip(popped_nodes_flat, states_c_by_node)]

    # Get is_solved on all states at once (for speed)
    states_c: List[State]

    states_c, split_idxs_c = misc_utils.flatten(states_c_by_node)
    goals_c, _ = misc_utils.flatten(goals_c_by_node)
    is_solved_c: List[bool] = list(env.is_solved(states_c, goals_c))
    is_solved_c_by_node: List[List[bool]] = misc_utils.unflatten(is_solved_c, split_idxs_c)

    # Update path costs for all states at once (for speed)
    # TODO, fix for variable action spaces
    parent_path_costs = np.expand_dims(np.array([node.path_cost for node in popped_nodes_flat]), 1)
    path_costs_c: List[float] = (parent_path_costs + np.array(tcs_by_node)).flatten().tolist()

    path_costs_c_by_node: List[List[float]] = misc_utils.unflatten(path_costs_c, split_idxs_c)

    # Reshape lists
    tcs_by_inst_node: List[List[List[float]]] = misc_utils.unflatten(tcs_by_node, split_idxs)
    patch_costs_c_by_inst_node: List[List[List[float]]] = misc_utils.unflatten(path_costs_c_by_node,
                                                                               split_idxs)
    states_c_by_inst_node: List[List[List[State]]] = misc_utils.unflatten(states_c_by_node, split_idxs)
    actions_c_by_inst_node: List[List[List[Action]]] = misc_utils.unflatten(actions_c_by_node, split_idxs)
    is_solved_c_by_inst_node: List[List[List[bool]]] = misc_utils.unflatten(is_solved_c_by_node, split_idxs)

    # Get child nodes
    instance: Instance
    nodes_c_by_inst: List[List[Node]] = []
    for inst_idx, instance in enumerate(instances):
        nodes_c_by_inst.append([])
        parent_nodes: List[Node] = popped_nodes_all[inst_idx]
        tcs_by_node = tcs_by_inst_node[inst_idx]
        path_costs_c_by_node = patch_costs_c_by_inst_node[inst_idx]
        states_c_by_node = states_c_by_inst_node[inst_idx]
        actions_c_by_node = actions_c_by_inst_node[inst_idx]

        is_solved_c_by_node = is_solved_c_by_inst_node[inst_idx]

        parent_node: Node
        tcs_node: List[float]
        for parent_node, tcs_node, path_costs_c, states_c, actions_c, is_solved_c in (zip(parent_nodes, tcs_by_node,
                                                                                          path_costs_c_by_node,
                                                                                          states_c_by_node,
                                                                                          actions_c_by_node,
                                                                                          is_solved_c_by_node)):
            state: State
            goal: Goal = parent_node.goal
            for move_idx, state in enumerate(states_c):
                path_cost: float = path_costs_c[move_idx]
                is_solved: bool = is_solved_c[move_idx]
                action: Action = actions_c[move_idx]
                node_c: Node = Node(state, goal, path_cost, is_solved, action, parent_node)

                nodes_c_by_inst[inst_idx].append(node_c)

        instance.num_nodes_generated += len(nodes_c_by_inst[inst_idx])

    return nodes_c_by_inst


def remove_in_closed(instances: List[Instance], nodes_c_all: List[List[Node]]) -> List[List[Node]]:
    for inst_idx, instance in enumerate(instances):
        nodes_c_all[inst_idx] = instance.remove_in_closed(nodes_c_all[inst_idx])

    return nodes_c_all


def add_to_open(instances: List[Instance], nodes: List[List[Node]]) -> None:
    nodes_inst: List[Node]
    instance: Instance
    for instance, nodes_inst in zip(instances, nodes):
        instance.push_to_open(nodes_inst)


def add_heuristic_and_cost(nodes: List[Node], heuristic_fn: HeurFN_T,
                           weights: List[float]) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    if len(nodes) == 0:
        return np.zeros(0), np.zeros(0)

    # compute node cost
    states: List[State] = [node.state for node in nodes]
    goals: List[Goal] = [node.goal for node in nodes]
    heuristics = heuristic_fn(states, goals)
    path_costs: NDArray[np.float64] = np.array([node.path_cost for node in nodes])
    is_solved: NDArray[np.bool_] = np.array([node.is_solved for node in nodes])

    costs: NDArray[np.float64] = np.array(weights) * path_costs + heuristics * np.logical_not(is_solved)

    # add cost to node
    for node, heuristic, cost in zip(nodes, heuristics, costs):
        node.heuristic = heuristic
        node.cost = cost

    return path_costs, heuristics


class AStar:

    def __init__(self, env: Environment):
        """ Initialize AStar search

        :param env: Environment
        """
        self.env: Environment = env
        self.step_num: int = 0

        self.timings: Dict[str, float] = {"pop": 0.0, "expand": 0.0, "check": 0.0, "heur": 0.0,
                                          "add": 0.0, "itr": 0.0}
        self.instances: List[Instance] = []

    def add_instances(self, states: List[State], goals: List[Goal], weights: List[float], heuristic_fn: HeurFN_T):
        """ Add instances

        :param states: start states
        :param goals: goals
        :param weights: weights for weighted A* search
        :param heuristic_fn: heuristic function
        :return:
        """
        assert len(states) == len(goals), "Number of states and goals should be the same"
        assert len(goals) == len(weights), "Number of weights given should be the same as number of instances"
        # compute starting costs
        root_nodes: List[Node] = []
        is_solved_states: NDArray[np.bool_] = np.array(self.env.is_solved(states, goals))
        for state, goal, is_solved in zip(states, goals, is_solved_states):
            root_node: Node = Node(state, goal, 0.0, is_solved, None, None)
            root_nodes.append(root_node)

        add_heuristic_and_cost(root_nodes, heuristic_fn, weights)

        # initialize instances
        for root_node, weight in zip(root_nodes, weights):
            self.instances.append(Instance(root_node, weight))

    def step(self, heuristic_fn: HeurFN_T, batch_size: int, verbose: bool = False):
        """ Take a step of A* search

        :param heuristic_fn: heuristic function
        :param batch_size: batch size
        :param verbose: If true, prints out search step information
        :return:
        """
        start_time_itr = time.time()
        instances: List[Instance] = [instance for instance in self.instances if not instance.finished]

        # Pop from open
        start_time = time.time()
        popped_nodes_all: List[List[Node]] = [instance.pop_from_open(batch_size) for instance in instances]
        pop_time = time.time() - start_time

        # Expand nodes
        start_time = time.time()
        nodes_c_all: List[List[Node]] = expand_nodes(instances, popped_nodes_all, self.env)
        expand_time = time.time() - start_time

        # Get heuristic of children, do heur before check so we can do backup
        start_time = time.time()
        nodes_c_all_flat, _ = misc_utils.flatten(nodes_c_all)
        weights, _ = misc_utils.flatten([[instance.weight] * len(nodes_c)
                                         for instance, nodes_c in zip(instances, nodes_c_all)])
        path_costs, heuristics = add_heuristic_and_cost(nodes_c_all_flat, heuristic_fn, weights)
        heur_time = time.time() - start_time

        # Check if children are in closed
        start_time = time.time()
        nodes_c_all = remove_in_closed(instances, nodes_c_all)
        check_time = time.time() - start_time

        # Add to open
        start_time = time.time()
        add_to_open(instances, nodes_c_all)
        add_time = time.time() - start_time

        itr_time = time.time() - start_time_itr

        # Print to screen
        if verbose:
            if heuristics.shape[0] > 0:
                min_heur = float(np.min(heuristics))
                min_heur_pc = float(path_costs[np.argmin(heuristics)])
                max_heur = float(np.max(heuristics))
                max_heur_pc = float(path_costs[np.argmax(heuristics)])

                print("Itr: %i, Added to OPEN - Min/Max Heur(PathCost): "
                      "%.2f(%.2f)/%.2f(%.2f) " % (self.step_num, min_heur, min_heur_pc, max_heur, max_heur_pc))

            print("Times - pop: %.2f, expand: %.2f, check: %.2f, heur: %.2f, "
                  "add: %.2f, itr: %.2f" % (pop_time, expand_time, check_time, heur_time, add_time, itr_time))

            print("")

        # Update timings
        self.timings['pop'] += pop_time
        self.timings['expand'] += expand_time
        self.timings['check'] += check_time
        self.timings['heur'] += heur_time
        self.timings['add'] += add_time
        self.timings['itr'] += itr_time

        self.step_num += 1
        for instance in instances:
            instance.step_num += 1

    def remove_instances(self, test_rem: Callable[[Instance], bool]) -> List[Instance]:
        """ Remove instances

        :param test_rem: A Callable that takes an instance as input and returns true if the instance should be removed
        :return: List of removed instances
        """
        instances_remove: List[Instance] = []
        instances_keep: List[Instance] = []
        for instance in self.instances:
            if test_rem(instance):
                instances_remove.append(instance)
            else:
                instances_keep.append(instance)

        self.instances = instances_keep

        return instances_remove


def get_path(node: Node) -> Tuple[List[State], List[Action], float]:
    """ Gets path from a the start state to the goal state associated with the input node

    :param node: goal node
    :return: List of states along path, List of actions in path, path cost
    """
    path: List[State] = []
    actions: List[Action] = []

    parent_node: Node = node
    while parent_node.parent is not None:
        path.append(parent_node.state)

        assert parent_node.parent_action is not None, "parent_action should not be None"
        actions.append(parent_node.parent_action)
        parent_node = parent_node.parent

    path.append(parent_node.state)

    path = path[::-1]
    actions = actions[::-1]

    return path, actions, node.path_cost



from typing import List, Tuple, Any
from deepxube.utils import data_utils
from deepxube.nnet import nnet_utils
from deepxube.nnet.nnet_utils import HeurFnQ
from deepxube.environments.environment_abstract import State, Environment, Goal
from deepxube.updaters.updater import Updater
from deepxube.search.greedy_policy import greedy_test

import torch
from torch import Tensor
import torch.optim as optim
from torch.optim.optimizer import Optimizer
import torch.nn as nn

import os
import pickle

import numpy as np
from numpy.typing import NDArray
import time

import sys
import shutil
import random


class Status:
    def __init__(self, env: Environment, step_max: int, num_test_per_step: int):
        self.itr: int = 0
        self.update_num: int = 0

        # generate data
        self.state_t_steps_l: List[int] = []
        for step in range(step_max + 1):
            self.state_t_steps_l.extend([step] * num_test_per_step)
        random.shuffle(self.state_t_steps_l)

        self.states_start_t: List[State]
        self.goals_t: List[Goal]
        print(f"Generating {num_test_per_step} test states per step ({step_max} steps, "
              f"{format(len(self.state_t_steps_l), ',')} total test states)")
        self.states_start_t, self.goals_t = env.get_start_goal_pairs(self.state_t_steps_l)

        # Initialize per_solved_best
        print("Initializing per solved best")
        heur_fn_qs, heur_procs = nnet_utils.start_heur_fn_runners(1, "", torch.device("cpu"), False, env.get_v_nnet(),
                                                                  env, all_zeros=True)
        per_solved: float = greedy_test(self.states_start_t, self.goals_t, self.state_t_steps_l, env,
                                        heur_fn_qs, max_solve_steps=1)
        print("Greedy policy solved: %.2f%%" % per_solved)
        self.per_solved_best: float = per_solved


def do_update(step_max: int, update_num: int, env: Environment, step_update_max: int, num_states: int,
              eps_max: float, heur_fn_qs: List[HeurFnQ],
              update_batch_size: int) -> Tuple[List[NDArray[Any]], NDArray[np.float64]]:
    update_steps: int = int(min(update_num + 1, step_update_max))
    # num_states: int = int(np.ceil(num_states / update_steps))

    if update_num == 0:
        eps_max = 1.0

    # Do updates
    output_time_start = time.time()

    print(f"Updating cost-to-go with value iteration. Generating {format(num_states, ',')} training states.")
    step_probs: List[float] = list(np.ones(1 + step_max) / (step_max + 1))
    if step_update_max > 1:
        print("Using greedy policy with %i step(s)" % update_steps)

    updater: Updater = Updater(env, num_states, step_max, step_probs, heur_fn_qs, update_steps,
                               "greedy", update_batch_size=update_batch_size, eps_max=eps_max)

    nnet_rep: List[NDArray[Any]]
    ctgs: NDArray[np.float64]
    nnet_rep, ctgs, is_solved = updater.update()

    # Print stats
    # per_solved = 100.0 * np.mean(is_solved)
    print("Produced %s states, (%.2f seconds)" % (format(ctgs.shape[0], ","), time.time() - output_time_start))

    mean_ctg = ctgs[:, 0].mean()
    min_ctg = ctgs[:, 0].min()
    max_ctg = ctgs[:, 0].max()
    print("Cost-to-go (mean/min/max): %.2f/%.2f/%.2f" % (mean_ctg, min_ctg, max_ctg))

    return nnet_rep, ctgs


def load_data(model_dir: str, nnet_file: str, env: Environment, num_test_per_step: int,
              step_max: int) -> Tuple[nn.Module, Status]:
    status_file: str = "%s/status.pkl" % model_dir
    if os.path.isfile(nnet_file):
        nnet = nnet_utils.load_nnet(nnet_file, env.get_v_nnet())
    else:
        nnet = env.get_v_nnet()

    status: Status
    if os.path.isfile(status_file):
        status = pickle.load(open("%s/status.pkl" % model_dir, "rb"))
        print(f"Loaded with itr: {status.itr}, update_num: {status.update_num}, "
              f"per_solved_best: {status.per_solved_best}")
    else:
        status = Status(env, step_max, num_test_per_step)
        pickle.dump(status, open(status_file, "wb"), protocol=-1)

    return nnet, status


def make_batches(nnet_rep: List[NDArray[Any]], ctgs: NDArray[np.float64],
                 batch_size: int) -> List[Tuple[List[NDArray[Any]], NDArray[np.float64]]]:
    num_examples = ctgs.shape[0]
    rand_idxs = np.random.choice(num_examples, num_examples, replace=False)
    ctgs = ctgs.astype(np.float32)

    start_idx = 0
    batches = []
    while (start_idx + batch_size) <= num_examples:
        end_idx = start_idx + batch_size

        idxs = rand_idxs[start_idx:end_idx]

        inputs_batch = [x[idxs] for x in nnet_rep]
        ctgs_batch = ctgs[idxs]

        batches.append((inputs_batch, ctgs_batch))

        start_idx = end_idx

    return batches


def train_nnet(nnet: nn.Module, nnet_rep: List[NDArray[Any]], ctgs: NDArray[np.float64], device: torch.device,
               batch_size: int, num_itrs: int, train_itr: int, lr: float, lr_d: float, display_itrs: int) -> float:
    # optimization
    criterion = nn.MSELoss()
    optimizer: Optimizer = optim.Adam(nnet.parameters(), lr=lr)

    # initialize status tracking
    start_time = time.time()

    # train network
    batches = make_batches(nnet_rep, ctgs, batch_size)

    nnet.train()
    max_itrs: int = train_itr + num_itrs

    last_loss: float = np.inf
    batch_idx: int = 0
    while train_itr < max_itrs:
        # zero the parameter gradients
        optimizer.zero_grad()
        lr_itr: float = lr * (lr_d ** train_itr)
        for param_group in optimizer.param_groups:
            param_group['lr'] = lr_itr

        # get data
        inputs_batch_np, ctgs_batch_np = batches[batch_idx]
        ctgs_batch_np = ctgs_batch_np.astype(np.float32)

        # send data to device
        inputs_batch: List[Tensor] = nnet_utils.to_pytorch_input(inputs_batch_np, device)
        ctgs_batch: Tensor = torch.tensor(ctgs_batch_np, device=device)[:, 0]

        # forward
        ctgs_nnet: Tensor = nnet(inputs_batch)

        # loss
        loss = criterion(ctgs_nnet[:, 0], ctgs_batch)

        # backwards
        loss.backward()

        # step
        optimizer.step()

        last_loss = loss.item()
        # display progress
        if (display_itrs > 0) and (train_itr % display_itrs == 0):
            print("Itr: %i, lr: %.2E, loss: %.2E, targ_ctg: %.2f, nnet_ctg: %.2f, "
                  "Time: %.2f" % (
                      train_itr, lr_itr, loss.item(), ctgs_batch.mean().item(),
                      ctgs_nnet.mean().item(), time.time() - start_time))

            start_time = time.time()

        train_itr = train_itr + 1

        batch_idx += 1
        if batch_idx >= len(batches):
            random.shuffle(batches)
            batch_idx = 0

    return last_loss


def train(env: Environment, step_max: int, nnet_dir: str, num_test_per_step: int = 30,
          itrs_per_update: int = 5000, epochs_per_update: int = 1, num_update_procs: int = 1,
          update_batch_size: int = 10000, update_nnet_batch_size: int = 10000, greedy_update_step_max: int = 1,
          greedy_update_eps_max: float = 0.1, lr: float = 0.001, lr_d: float = 0.9999993, max_itrs: int = 1000000,
          batch_size: int = 1000, display: int = 100, debug: bool = False):
    """ Train a deep neural network heuristic (DNN) function with deep approximate value iteration (DAVI).
    A target DNN is maintained for computing the updated heuristic values. When the greedy policy improves on a fixed
    test set, the target DNN is updated to be the current DNN. The number of steps taken for testing the greedy policy
    is the minimum between the number of target DNN updates and step_max.
    This makes the test a lot faster in the earlier stages, espeicially when step_max is large.

    For more information see:
    - Agostinelli, Forest, et al. "Solving the Rubik’s cube with deep reinforcement learning and search."
    Nature Machine Intelligence 1.8 (2019): 356-363.
    - Bertsekas, D. P. & Tsitsiklis, J. N. Neuro-dynamic Programming (Athena Scientific, 1996).

    :param env: an Environment object
    :param step_max: maximum number of steps to take to generate start/goal pairs
    :param nnet_dir: directory where DNN will be saved
    :param num_test_per_step: Number of test states for each step between 0 and step_max
    :param itrs_per_update: How many iterations to do before checking if target network should be updated
    :param epochs_per_update: How many epochs for which train. Making this greater than 1 could increase risk of
    overfitting, however, one can train for more iterations without having to generate more data.)
    :param num_update_procs: Number of parallel workers used to compute updated cost-to-go values
    :param update_batch_size: Maximum number of start/goal pairs when computing updated cost-to-go that each
    parallel worker generates at a time. Since multiprocessing is used, memory issues can happen if trying to send
    objects that are too large across processes. Therefore, if you update step freezes for no apparent reason,
    it could be a memory issue. Lowering this value or making the DNN representation more memory efficient could solve
    it.
    :param update_nnet_batch_size: Batch size of each nnet used for each process update. Make smaller if running out
    of memory.
    :param greedy_update_step_max: Maximum number of epsilon greedy policy steps (update_steps) to take from generated
    start states to generate additional data. Value of 1 is the same as basic DAVI. Increasing this number could make
    the heuristic function more robust to depression regions. The number of steps taken for the update is the
    minimum between the number of target DNN updates and greedy_update_step_max. Number of start/goal pairs is
    states_per_update/update_steps.
    :param greedy_update_eps_max: epsilon greedy policy max. Each start/goal pair will have an eps uniformly distributed
    between 0 and greedy_update_eps_max
    :param lr: Initial learning rate
    :param lr_d: Learning rate decay for every iteration. Learning rate is decayed according to: lr * (lr_d ^ itr)
    :param max_itrs: Maximum number of iterations
    :param batch_size: Batch size
    :param display: Number of iterations to display progress. No display if 0.
    :param debug: Turns off logging to make typing during breakpoints easier
    :return: None
    """
    # Initialization
    targ_file: str = f"{nnet_dir}/target.pt"
    curr_file = f"{nnet_dir}/current.pt"
    output_save_loc = "%s/output.txt" % nnet_dir

    if not os.path.exists(nnet_dir):
        os.makedirs(nnet_dir)

    if not debug:
        sys.stdout = data_utils.Logger(output_save_loc, "a")  # type: ignore

    # Print basic info
    # print("HOST: %s" % os.uname()[1])
    print("CPU: %i" % num_update_procs)
    print("Batch size: %i" % batch_size)
    if 'SLURM_JOB_ID' in os.environ:
        print("SLURM JOB ID: %s" % os.environ['SLURM_JOB_ID'])

    # get device
    on_gpu: bool
    device: torch.device
    device, devices, on_gpu = nnet_utils.get_device()

    print("device: %s, devices: %s, on_gpu: %s" % (device, devices, on_gpu))

    # load nnet
    print("Loading nnet and status")
    nnet, status = load_data(nnet_dir, curr_file, env, num_test_per_step, step_max)
    nnet.to(device)
    nnet = nn.DataParallel(nnet)

    # training
    while status.itr < max_itrs:
        # update
        all_zeros: bool = not os.path.isfile(targ_file)
        heur_fn_qs, heur_procs = nnet_utils.start_heur_fn_runners(num_update_procs, targ_file,
                                                                  device, on_gpu, env.get_v_nnet(), env,
                                                                  all_zeros=all_zeros, clip_zero=True,
                                                                  batch_size=update_nnet_batch_size)

        states_per_update: int = itrs_per_update * batch_size
        num_update_states: int = int(states_per_update)

        nnet_rep, ctgs = do_update(step_max, status.update_num, env, greedy_update_step_max, num_update_states,
                                   greedy_update_eps_max, heur_fn_qs, update_batch_size)

        nnet_utils.stop_heuristic_fn_runners(heur_procs, heur_fn_qs)

        # train nnet
        num_train_itrs: int = epochs_per_update * np.ceil(ctgs.shape[0] / batch_size)
        print("Training model for update number %i for %i iterations" % (status.update_num, num_train_itrs))
        last_loss = train_nnet(nnet, nnet_rep, ctgs, device, batch_size, num_train_itrs,
                               status.itr, lr, lr_d, display)
        status.itr += num_train_itrs

        # save nnet
        torch.save(nnet.state_dict(), curr_file)

        # test
        start_time = time.time()

        heur_fn_qs, heur_procs = nnet_utils.start_heur_fn_runners(num_update_procs, curr_file,
                                                                  device, on_gpu, env.get_v_nnet(), env,
                                                                  all_zeros=False, clip_zero=False,
                                                                  batch_size=update_nnet_batch_size)

        max_solve_steps: int = min(status.update_num + 1, step_max)

        print("Testing greedy policy with %i states and %i steps" % (len(status.states_start_t), max_solve_steps))
        per_solved: float = greedy_test(status.states_start_t, status.goals_t, status.state_t_steps_l, env,
                                        heur_fn_qs, max_solve_steps=max_solve_steps)
        print("Greedy policy solved (best): %.2f%% (%.2f%%)" % (per_solved, status.per_solved_best))

        nnet_utils.stop_heuristic_fn_runners(heur_procs, heur_fn_qs)
        print("Test time: %.2f" % (time.time() - start_time))

        # clear cuda memory
        torch.cuda.empty_cache()

        update_nnet: bool
        if per_solved > status.per_solved_best:
            print("Updating target network")
            status.per_solved_best = per_solved
            update_nnet = True
        else:
            update_nnet = False

        print("Last loss was %f" % last_loss)
        if update_nnet:
            # Update nnet
            shutil.copy(curr_file, targ_file)
            status.update_num = status.update_num + 1

        pickle.dump(status, open("%s/status.pkl" % nnet_dir, "wb"), protocol=-1)

    print("Done")



from typing import List, Tuple, Optional, Callable, Any, Union, cast
from deepxube.environments.environment_abstract import Environment, State, Goal, HeurFnNNet
import numpy as np
from numpy.typing import NDArray
import os
import torch
from torch import nn
from collections import OrderedDict
import re
from torch import Tensor
from torch.multiprocessing import Queue, get_context
from multiprocessing.process import BaseProcess

HeurFN_T = Callable[[Union[List[State], List[NDArray[Any]]], Optional[List[Goal]]], NDArray[np.float64]]


# training
def to_pytorch_input(states_nnet: List[NDArray[Any]], device) -> List[Tensor]:
    states_nnet_tensors = []
    for tensor_np in states_nnet:
        tensor = torch.tensor(tensor_np, device=device)
        states_nnet_tensors.append(tensor)

    return states_nnet_tensors


# pytorch device
def get_device() -> Tuple[torch.device, List[int], bool]:
    device: torch.device = torch.device("cpu")
    devices: List[int] = []
    on_gpu: bool = False
    if ('CUDA_VISIBLE_DEVICES' in os.environ) and torch.cuda.is_available():
        device = torch.device("cuda:%i" % 0)
        devices = [int(x) for x in os.environ['CUDA_VISIBLE_DEVICES'].split(",")]
        on_gpu = True
    else:
        torch.set_num_threads(8)

    return device, devices, on_gpu


# loading nnet
def load_nnet(model_file: str, nnet: nn.Module, device: Optional[torch.device] = None) -> nn.Module:
    # get state dict
    if device is None:
        state_dict = torch.load(model_file)
    else:
        state_dict = torch.load(model_file, map_location=device)

    # remove module prefix
    new_state_dict = OrderedDict()
    for k, v in state_dict.items():
        k = re.sub(r'^module\.', '', k)
        new_state_dict[k] = v

    # set state dict
    nnet.load_state_dict(new_state_dict)

    nnet.eval()

    return nnet


# heuristic
def get_heuristic_fn(nnet: nn.Module, device: torch.device, env: Environment, clip_zero: bool = False,
                     batch_size: Optional[int] = None, is_v: bool = False) -> HeurFN_T:
    nnet.eval()

    def heuristic_fn(states: Union[List[State], List[NDArray[Any]]],
                     goals: Optional[List[Goal]]) -> NDArray[np.float64]:
        cost_to_go_l: List[NDArray[np.float64]] = []

        num_states: int
        is_nnet_format: bool
        if goals is not None:
            num_states = len(states)
            is_nnet_format = False
        else:
            num_states = cast(List[NDArray[Any]], states)[0].shape[0]
            is_nnet_format = True

        batch_size_inst: int = num_states
        if batch_size is not None:
            batch_size_inst = batch_size

        start_idx: int = 0
        while start_idx < num_states:
            # get batch
            end_idx: int = min(start_idx + batch_size_inst, num_states)

            # convert to nnet input
            if not is_nnet_format:
                states_batch: List[State] = cast(List[State], states)[start_idx:end_idx]
                goals_batch: List[Goal] = cast(List[Goal], goals)[start_idx:end_idx]

                states_goals_nnet_batch = env.states_goals_to_nnet_input(states_batch, goals_batch)
            else:
                states_goals_nnet_batch = [x[start_idx:end_idx] for x in cast(List[NDArray[Any]], states)]

            # get nnet output
            states_goals_nnet_batch_tensors = to_pytorch_input(states_goals_nnet_batch, device)

            cost_to_go_batch: NDArray[np.float64] = nnet(states_goals_nnet_batch_tensors).cpu().data.numpy()
            if is_v:
                cost_to_go_batch = cost_to_go_batch[:, 0]
            cost_to_go_l.append(cost_to_go_batch)

            start_idx = end_idx

        cost_to_go = np.concatenate(cost_to_go_l, axis=0)
        assert (cost_to_go.shape[0] == num_states)

        if clip_zero:
            cost_to_go = np.maximum(cost_to_go, 0.0)

        return cost_to_go

    return heuristic_fn


def get_available_gpu_nums() -> List[int]:
    gpu_nums: List[int] = []
    if ('CUDA_VISIBLE_DEVICES' in os.environ) and (len(os.environ['CUDA_VISIBLE_DEVICES']) > 0):
        gpu_nums = [int(x) for x in os.environ['CUDA_VISIBLE_DEVICES'].split(",")]

    return gpu_nums


def load_heuristic_fn(model_file: str, device: torch.device, on_gpu: bool, nnet: nn.Module,
                      env: Environment, clip_zero: bool = False, gpu_num: Optional[int] = None,
                      batch_size: Optional[int] = None) -> HeurFN_T:
    if (gpu_num is not None) and on_gpu:
        os.environ['CUDA_VISIBLE_DEVICES'] = str(gpu_num)

    is_v: bool = nnet.nnet_type.upper() == "V"

    nnet = load_nnet(model_file, nnet, device=device)
    nnet.eval()
    nnet.to(device)
    if on_gpu:
        nnet = nn.DataParallel(nnet)

    heuristic_fn = get_heuristic_fn(nnet, device, env, clip_zero=clip_zero, batch_size=batch_size, is_v=is_v)

    return heuristic_fn


# parallel training
def heuristic_fn_runner(heuristic_fn_input_queue: Queue, heuristic_fn_output_queues: List[Queue],
                        model_file: str, device, on_gpu: bool, gpu_num: int, nnet: HeurFnNNet,
                        env: Environment, all_zeros: bool, clip_zero: bool, batch_size: Optional[int]):
    heuristic_fn: Optional[HeurFN_T] = None
    if not all_zeros:
        heuristic_fn = load_heuristic_fn(model_file, device, on_gpu, nnet, env, gpu_num=gpu_num,
                                         clip_zero=clip_zero, batch_size=batch_size)

    while True:
        proc_id, states_goals_nnet = heuristic_fn_input_queue.get()
        if proc_id is None:
            break

        if all_zeros:
            heuristics = np.zeros(states_goals_nnet[0].shape[0], dtype=float)
        else:
            heuristics = cast(HeurFN_T, heuristic_fn)(states_goals_nnet, None)

        heuristic_fn_output_queues[proc_id].put(heuristics)

    return heuristic_fn


class HeurFnQ:
    def __init__(self, heur_fn_i_q, heur_fn_o_q, proc_id: int):
        self.heur_fn_i_q = heur_fn_i_q
        self.heur_fn_o_q = heur_fn_o_q
        self.proc_id: int = proc_id

    def get_heuristic_fn(self, env: Environment) -> HeurFN_T:
        def heuristic_fn(states: Any, goals: Optional[List[Goal]]):
            # states: Union[List[State], NDArray[Any]]
            if goals is not None:
                states_goals_nnet = env.states_goals_to_nnet_input(states, goals)
            else:
                states_goals_nnet = states
            self.heur_fn_i_q.put((self.proc_id, states_goals_nnet))

            heuristics = self.heur_fn_o_q.get()

            return heuristics

        return heuristic_fn


def start_heur_fn_runners(num_procs: int, model_file: str, device, on_gpu: bool, nnet: HeurFnNNet,
                          env: Environment, all_zeros: bool = False, clip_zero: bool = False,
                          batch_size: Optional[int] = None) -> Tuple[List[HeurFnQ], List[BaseProcess]]:
    ctx = get_context("spawn")

    heur_fn_i_q: Queue = ctx.Queue()
    heur_fn_o_qs: List[Queue] = []
    heur_fn_qs: List[HeurFnQ] = []
    for proc_id in range(num_procs):
        heur_fn_o_q: Queue = ctx.Queue(1)
        heur_fn_o_qs.append(heur_fn_o_q)
        heur_fn_qs.append(HeurFnQ(heur_fn_i_q, heur_fn_o_q, proc_id))

    # initialize heuristic procs
    if ('CUDA_VISIBLE_DEVICES' in os.environ) and (len(os.environ['CUDA_VISIBLE_DEVICES']) > 0):
        gpu_nums = [int(x) for x in os.environ['CUDA_VISIBLE_DEVICES'].split(",")]
    else:
        gpu_nums = [-1]

    heur_procs: List[BaseProcess] = []
    for gpu_num in gpu_nums:
        heur_fn_proc = ctx.Process(target=heuristic_fn_runner,
                                   args=(heur_fn_i_q, heur_fn_o_qs, model_file, device, on_gpu, gpu_num, nnet,
                                         env, all_zeros, clip_zero, batch_size))
        heur_fn_proc.daemon = True
        heur_fn_proc.start()
        heur_procs.append(heur_fn_proc)

    return heur_fn_qs, heur_procs


def stop_heuristic_fn_runners(heur_fn_procs, heur_fn_qs: List[HeurFnQ]):
    for _ in heur_fn_procs:
        heur_fn_qs[0].heur_fn_i_q.put((None, None))

    for heur_fn_proc in heur_fn_procs:
        heur_fn_proc.join()



from typing import List, Tuple, Set, Callable, Optional
from deepxube.environments.environment_abstract import Environment, State, Goal
from deepxube.utils import misc_utils
from deepxube.utils.timing_utils import Times
from deepxube.nnet.nnet_utils import HeurFnQ, HeurFN_T
import numpy as np
from numpy.typing import NDArray
from deepxube.search.search_utils import bellman
from torch.multiprocessing import get_context, Queue
import random
import time


class Instance:

    def __init__(self, state: State, goal: Goal, eps: float):
        self.curr_state: State = state
        self.is_solved: bool = False
        self.goal: Goal = goal
        self.step_num: int = 0
        self.traj: List[Tuple[State, float]] = []
        self.seen_states: Set[State] = set()

        self.eps = eps

    def add_to_traj(self, state: State, cost_to_go: float):
        self.traj.append((state, cost_to_go))
        self.seen_states.add(state)

    def next_state(self, state: State):
        self.curr_state = state
        self.step_num += 1


class Greedy:
    def __init__(self, env: Environment):
        self.env: Environment = env
        self.instances: List[Instance] = []

    def add_instances(self, states: List[State], goals: List[Goal], eps_l: Optional[List[float]]):
        if eps_l is None:
            eps_l = [0] * len(states)

        assert len(states) == len(goals), "Number of states and goals should be the same"
        assert len(goals) == len(eps_l), "Number of epsilon given should be the same as number of instances"

        for state, goal, eps_inst in zip(states, goals, eps_l):
            instance: Instance = Instance(state, goal, eps_inst)
            self.instances.append(instance)

    def step(self, heuristic_fn: HeurFN_T, times: Optional[Times] = None,
             rand_seen: bool = False) -> Tuple[List[State], List[Goal], NDArray[np.float64]]:
        if times is None:
            times = Times()

        # get unsolved instances
        start_time = time.time()
        instances: List[Instance] = self._get_unsolved_instances()
        if len(instances) == 0:
            return [], [], np.zeros(0)

        states: List[State] = [instance.curr_state for instance in instances]
        goals: List[Goal] = [instance.goal for instance in instances]
        times.record_time("get_unsolved", time.time() - start_time)

        # bellman
        ctg_backups, ctg_next_p_tcs, states_exp, is_solved = bellman(states, goals, heuristic_fn, self.env, times)

        # take action
        start_time = time.time()
        rand_vals = np.random.random(len(instances))
        for idx in range(len(instances)):
            # add state to trajectory
            instance: Instance = instances[idx]
            state: State = states[idx]
            ctg_backup: float = float(ctg_backups[idx])

            instance.add_to_traj(state, ctg_backup)

            # get next state
            if not is_solved[idx]:
                state_exp: List[State] = states_exp[idx]
                ctg_next_p_tc: NDArray[np.float64] = ctg_next_p_tcs[idx]

                state_next: State = state_exp[int(np.argmin(ctg_next_p_tc))]
                seen_state: bool = state_next in instance.seen_states
                if (rand_vals[idx] < instance.eps) or (seen_state and rand_seen):
                    state_next = random.choice(state_exp)
                instance.next_state(state_next)
        times.record_time("get_next", time.time() - start_time)

        # check which are solved
        start_time = time.time()
        states_next: List[State] = [instance.curr_state for instance in instances]
        goals = [instance.goal for instance in instances]
        is_solved_next: List[bool] = self.env.is_solved(states_next, goals)
        solved_idxs: List[int] = list(np.where(is_solved_next)[0])
        for solved_idx in solved_idxs:
            instances[solved_idx].is_solved = True
        times.record_time("record_solved", time.time() - start_time)

        return states, goals, ctg_backups

    def remove_instances(self, test_rem: Callable[[Instance], bool]) -> List[Instance]:
        """ Remove instances

        :param test_rem: A Callable that takes an instance as input and returns true if the instance should be removed
        :return: List of removed instances
        """
        instances_remove: List[Instance] = []
        instances_keep: List[Instance] = []
        for instance in self.instances:
            if test_rem(instance):
                instances_remove.append(instance)
            else:
                instances_keep.append(instance)

        self.instances = instances_keep

        return instances_remove

    def _get_unsolved_instances(self) -> List[Instance]:
        instances_unsolved: List[Instance] = [instance for instance in self.instances if not instance.is_solved]
        return instances_unsolved


def greedy_runner(env: Environment, heur_fn_q: HeurFnQ, proc_id: int,
                  max_solve_steps: int, data_q, results_queue):
    heuristic_fn = heur_fn_q.get_heuristic_fn(env)
    states, goals, inst_gen_steps = data_q.get()

    # Solve with GBFS
    greedy = Greedy(env)
    greedy.add_instances(states, goals, eps_l=None)
    for _ in range(max_solve_steps):
        greedy.step(heuristic_fn, rand_seen=True)

    is_solved_all: NDArray[np.bool_] = np.array([instance.is_solved for instance in greedy.instances])
    num_steps_all: NDArray[np.int_] = np.array([instance.step_num for instance in greedy.instances])

    # Get state cost-to-go
    state_ctg_all: NDArray[np.float64] = heuristic_fn(states, goals)

    results_queue.put((proc_id, is_solved_all, num_steps_all, state_ctg_all, inst_gen_steps))


def greedy_test(states: List[State], goals: List[Goal], inst_gen_steps: List[int], env: Environment,
                heur_fn_qs: List[HeurFnQ], max_solve_steps: Optional[int] = None) -> float:
    # initialize
    if max_solve_steps is None:
        max_solve_steps = max(max(inst_gen_steps), 1)

    ctx = get_context("spawn")
    data_q: Queue = ctx.Queue()
    results_q: Queue = ctx.Queue()
    procs: List = []
    num_states_per_proc: List[int] = misc_utils.split_evenly(len(states), len(heur_fn_qs))
    start_idx: int = 0
    for proc_id, heur_fn_q in enumerate(heur_fn_qs):
        num_states_proc: int = num_states_per_proc[proc_id]
        if num_states_proc == 0:
            continue
        # update_runner(num_states_proc, back_max, step_probs, update_batch_size,
        #              heur_fn_q, env.env_name, self.result_queue, solve_steps, update_method, eps_max)

        # start process
        end_idx: int = start_idx + num_states_proc
        proc = ctx.Process(target=greedy_runner, args=(env, heur_fn_q, proc_id, max_solve_steps, data_q, results_q))
        proc.daemon = True
        proc.start()
        procs.append(proc)

        # put data
        states_proc = states[start_idx:end_idx]
        goals_proc = goals[start_idx:end_idx]
        inst_gen_steps_proc = inst_gen_steps[start_idx:end_idx]
        data_q.put((states_proc, goals_proc, inst_gen_steps_proc))
        start_idx = end_idx

    is_solved_l: List[List[bool]] = [[] for _ in heur_fn_qs]
    num_steps_l: List[List[int]] = [[] for _ in heur_fn_qs]
    ctgs_l: List[List[float]] = [[] for _ in heur_fn_qs]
    inst_gen_steps_l: List[List[int]] = [[] for _ in heur_fn_qs]

    for _ in heur_fn_qs:
        proc_id, is_solved_i, num_steps_i, ctgs_i, inst_gen_steps_i = results_q.get()
        is_solved_l[proc_id] = is_solved_i
        num_steps_l[proc_id] = num_steps_i
        ctgs_l[proc_id] = ctgs_i
        inst_gen_steps_l[proc_id] = inst_gen_steps_i
    is_solved_all = np.hstack(is_solved_l)
    num_steps_all = np.hstack(num_steps_l)
    ctgs_all = np.hstack(ctgs_l)
    inst_gen_steps = list(np.hstack(inst_gen_steps_l))

    for proc in procs:
        proc.join()

    per_solved_all = 100 * float(sum(is_solved_all)) / float(len(is_solved_all))
    steps_show: List[int] = list(np.unique(np.linspace(0, max(inst_gen_steps), 30, dtype=int)))
    for back_step_test in np.sort(np.unique(steps_show)):
        # Get states
        step_idxs = np.where(np.array(inst_gen_steps) == back_step_test)[0]
        if len(step_idxs) == 0:
            continue

        is_solved: NDArray[np.bool_] = is_solved_all[step_idxs]
        num_steps: NDArray[np.int_] = num_steps_all[step_idxs]
        ctgs: NDArray[np.float64] = ctgs_all[step_idxs]

        # Get stats
        per_solved = 100 * float(sum(is_solved)) / float(len(is_solved))
        avg_solve_steps: float = 0.0
        if per_solved > 0.0:
            avg_solve_steps = float(np.mean(num_steps[is_solved]))

        # Print results
        print("Back Steps: %i, %%Solved: %.2f, avgSolveSteps: %.2f, CTG Mean(Std/Min/Max): %.2f("
              "%.2f/%.2f/%.2f)" % (
                  back_step_test, per_solved, avg_solve_steps, float(np.mean(ctgs)),
                  float(np.std(ctgs)), float(np.min(ctgs)), float(np.max(ctgs))))

    return per_solved_all



from typing import Optional, List, Tuple, Any
from deepxube.search.astar import AStar, Node, get_path
from deepxube.environments.environment_abstract import EnvGrndAtoms, State, Goal
from deepxube.nnet.nnet_utils import HeurFN_T
from deepxube.logic.logic_objects import Literal, Clause, Model, Atom
from deepxube.logic.logic_utils import atom_to_str
from deepxube.logic.asp import Solver, Spec
from deepxube.utils import viz_utils, misc_utils
from deepxube.utils.timing_utils import Times
import time


def get_bk(env: EnvGrndAtoms, bk_add_file_name: Optional[str]) -> List[str]:
    bk_init: List[str] = env.get_bk()
    bk_init.append("")

    if bk_add_file_name is not None:
        bk_add_file = open(bk_add_file_name, 'r')
        bk_init.extend(bk_add_file.read().split("\n"))
        bk_add_file.close()

    return bk_init


def search_for_goal(env: EnvGrndAtoms, state_start: State, models: List[Model], batch_size: int,
                    weight: float, max_search_itrs: int, heur_fn: HeurFN_T, search_verbose: bool,
                    viz_model: bool) -> List[Optional[Node]]:
    goals: List[Goal] = env.model_to_goal(models)

    # Check for None
    if len(models) == 0:
        return []

    # visualize
    if viz_model:
        print("Sampled goal visualization")
        viz_utils.visualize_examples(env, goals)

    # Initialize
    astar = AStar(env)
    astar.add_instances([state_start] * len(goals), goals, [weight] * len(goals), heur_fn)

    # Search
    search_itr: int = 0
    while (not min(x.finished for x in astar.instances)) and (search_itr < max_search_itrs):
        search_itr += 1
        astar.step(heur_fn, batch_size, verbose=search_verbose)

    return [x.goal_node for x in astar.instances]


def get_next_model(asp: Solver, spec_clauses: List[Clause], env: EnvGrndAtoms, models_banned: List[Model],
                   num_models: int, assumed_true: Optional[Model] = None,
                   num_atoms_gt: Optional[int] = None) -> List[Model]:
    if num_atoms_gt is not None:
        spec_clauses_new = []
        lit_count_gt: Literal = Literal("count_model_grnd_atoms_gt", (str(num_atoms_gt),), ("in",))
        for clause in spec_clauses:
            clause_new = Clause(clause.head, clause.body + (lit_count_gt,))
            spec_clauses_new.append(clause_new)
        spec_clauses = spec_clauses_new

    atoms_true: List[Atom] = []
    if assumed_true is not None:
        atoms_true = list(assumed_true)
    spec: Spec = Spec(goal_true=spec_clauses, atoms_true=atoms_true, models_banned=models_banned)
    models: List[Model] = asp.get_models(spec, env.on_model, num_models, True)

    return models


def path_to_spec_goal(env: EnvGrndAtoms, state_start: State, spec_clauses: List[Clause], heur_fn: HeurFN_T,
                      model_batch_size: int, search_batch_size: int, weight: float, max_search_itrs: int,
                      bk_add: Optional[str] = None, models_banned: Optional[List[Model]] = None,
                      times: Optional[Times] = None, spec_verbose: bool = False, search_verbose: bool = False,
                      viz_model: bool = False) -> Tuple[bool, List[State], List[Any], float, int, int, List[Model],
                                                        Times]:
    """

    :param env: EnvGrndAtoms environment
    :param state_start: Starting state
    :param spec_clauses: Clauses for specification. All must have goal in the head.
    :param heur_fn: Heuristic function
    :param model_batch_size: Maximum number of models sampled at once for parallelized search.
    :param search_batch_size: Batch size of search
    :param weight: Weight on path cost for weighted search. Must be between 0 and 1.
    :param max_search_itrs: Maximum number of iterations when searching from start state to goal model
    :param bk_add: A file for additional background information
    :param models_banned: Banned models
    :param times: Times
    :param spec_verbose: Verbose specification if true
    :param search_verbose: Verbose search if true
    :param viz_model: Set true to visualize each model before search
    :return: boolean that is true if a goal is found, list of states along path, list of actions along path, path cost,
    times
    """
    # Init
    if times is None:
        times = Times(time_names=["ASP init", "Model samp", "Search", "Check", "Model superset"])

    if models_banned is None:
        models_banned = []
    else:
        models_banned = models_banned.copy()

    # Initialize ASP
    start_time = time.time()
    model_fixed: Model = env.start_state_fixed([state_start])[0]
    bk: List[str] = get_bk(env, bk_add)
    bk += [atom_to_str(x) for x in model_fixed]
    asp: Solver = Solver(env.get_ground_atoms(), bk)
    times.record_time("ASP init", time.time() - start_time)

    # Sample initial models
    start_time = time.time()
    spec: Spec = Spec(goal_true=spec_clauses, atoms_true=list(model_fixed), models_banned=models_banned)
    models: List[Model] = asp.get_models(spec, env.on_model, model_batch_size, True)
    times.record_time("Model samp", time.time() - start_time)

    num_models_init: int = len(models)
    num_models_superset: int = 0
    while len(models) > 0:
        if spec_verbose:
            print(f"{len(models)} models")

        # search for goal nodes
        start_time = time.time()
        goal_nodes: List[Optional[Node]] = search_for_goal(env, state_start, models, search_batch_size, weight,
                                                           max_search_itrs, heur_fn, search_verbose, viz_model)

        models_banned += [model for model, goal_node in zip(models, goal_nodes) if goal_node is None]

        models = [model for model, goal_node in zip(models, goal_nodes) if goal_node is not None]
        goal_nodes_found: List[Node] = [goal_node for goal_node in goal_nodes if goal_node is not None]
        if spec_verbose:
            print(f"Found a path to {len(goal_nodes_found)} models")
        times.record_time("Search", time.time() - start_time)

        if len(models) == 0:
            continue

        # check for goal states
        models_terminal: List[Model] = env.state_to_model([goal_node.state for goal_node in goal_nodes_found])
        for goal_node, model_terminal in zip(goal_nodes_found, models_terminal):
            start_time = time.time()
            is_model = asp.check_model(Spec(goal_true=spec_clauses), model_terminal)
            times.record_time("Check", time.time() - start_time)
            if is_model:
                if spec_verbose:
                    print("Found a goal state")
                path_states, path_actions, path_cost = get_path(goal_node)
                return (True, path_states, path_actions, path_cost, num_models_init, num_models_superset, models_banned,
                        times)

        # Get supersets of models
        start_time = time.time()
        samp_per_model: List[int] = misc_utils.split_evenly(model_batch_size, len(models))
        models_superset: List[Model] = []
        for num_samp_i, model in zip(samp_per_model, models):
            spec_sup: Spec = Spec(goal_true=spec_clauses, atoms_true=list(model), models_banned=models_banned,
                                  num_atoms_gt=len(model))
            models_superset += asp.get_models(spec_sup, env.on_model, num_samp_i, True)

        num_models_superset += len(models_superset)
        models = models_superset
        times.record_time("Model superset", time.time() - start_time)

    if spec_verbose:
        print("No path found")

    return False, [], [], -1.0, num_models_init, num_models_superset, models_banned, times



from typing import List, Dict, Optional
from collections import OrderedDict


def add_times(times: OrderedDict[str, float], times_to_add: OrderedDict[str, float]):
    for key, value in times_to_add.items():
        if key not in times:
            times[key] = 0.0
        times[key] += value


def add_counts(counts: OrderedDict[str, int], counts_to_add: OrderedDict[str, int]):
    for key, value in counts_to_add.items():
        if key not in counts:
            counts[key] = 0
        counts[key] += value


def init_times(time_names: List[str]) -> OrderedDict[str, float]:
    times: OrderedDict[str, float] = OrderedDict()
    for time_name in time_names:
        times[time_name] = 0.0
    return times


def init_counts(time_names: List[str]) -> OrderedDict[str, int]:
    counts: OrderedDict[str, int] = OrderedDict()
    for time_name in time_names:
        counts[time_name] = 0
    return counts


class Times:

    def __init__(self, time_names: Optional[List[str]] = None):
        if time_names is None:
            time_names = []

        self.times: OrderedDict[str, float] = init_times(time_names)
        self.counts: OrderedDict[str, int] = init_counts(time_names)

        self.sub_times: Dict[str, Times] = dict()
        self.sub_counts: Dict[str, int] = dict()

    def record_time(self, time_name: str, time_elapsed: float, path: Optional[List[str]] = None):
        if (path is not None) and (len(path) > 0):
            path_0: str = path.pop(0)
            if path_0 not in self.sub_times:
                self.sub_times[path_0] = Times()
                self.sub_counts[path_0] = 0

            self.sub_times[path_0].record_time(time_name, time_elapsed, path=path)
            self.sub_counts[path_0] += 1
        else:
            if time_name not in self.times.keys():
                self.times[time_name] = 0
                self.counts[time_name] = 0

            self.times[time_name] += time_elapsed
            self.counts[time_name] += 1

    def add_times(self, time: 'Times', path: Optional[List[str]] = None):
        if (path is not None) and (len(path) > 0):
            path_0: str = path.pop(0)
            if path_0 not in self.sub_times:
                self.sub_times[path_0] = Times()
                self.sub_counts[path_0] = 0
            self.sub_times[path_0].add_times(time, path=path)
            self.sub_counts[path_0] += sum(time.counts.values())
        else:
            add_times(self.times, time.times)
            add_counts(self.counts, time.counts)
            for sub_time_name in time.sub_times.keys():
                if sub_time_name not in self.sub_times.keys():
                    self.sub_times[sub_time_name] = Times()
                    self.sub_counts[sub_time_name] = 0

                self.sub_times[sub_time_name].add_times(time.sub_times[sub_time_name])
                self.sub_counts[sub_time_name] += sum(time.sub_times[sub_time_name].counts.values())

    def reset_times(self):
        for key in self.times.keys():
            self.times[key] = 0.0
            self.counts[key] = 0

        for sub_time in self.sub_times.values():
            sub_time.reset_times()

    def get_total_time(self) -> float:
        time_tot: float = 0.0
        for time_elapsed in self.times.values():
            time_tot += time_elapsed

        for sub_time in self.sub_times.values():
            time_tot += sub_time.get_total_time()

        return time_tot

    def get_time_str(self, prefix: str = "") -> str:
        time_str_l: List[str] = ["%s: %f" % (key, val) for key, val in self.times.items()]
        sub_time_str_l: List[str] = ["->%s: %f" % (key, sub_time.get_total_time())
                                     for key, sub_time in self.sub_times.items()]

        time_str: str = ", ".join(time_str_l + sub_time_str_l + ["Tot: %f" % self.get_total_time()])

        prefix_new: str = f"\t{prefix}"
        for key, sub_time in self.sub_times.items():
            time_str = f"{time_str}\n{prefix_new}({key}): {sub_time.get_time_str(prefix=prefix_new)}"

        return time_str

    def __str__(self):
        return self.get_time_str()

    def __repr__(self):
        return self.__str__()



from typing import List, Tuple, Any
import numpy as np
from numpy.typing import NDArray
from deepxube.utils import misc_utils
from deepxube.nnet.nnet_utils import HeurFnQ
from deepxube.utils.timing_utils import Times
from deepxube.environments.environment_abstract import Environment, State, Goal
from deepxube.search.greedy_policy import Greedy
from torch.multiprocessing import get_context, Queue
from multiprocessing.process import BaseProcess
import time


def greedy_update(states: List[State], goals: List[Goal], env: Environment, num_steps: int, heuristic_fn,
                  eps_max: float, times: Times):
    eps: List[float] = list(np.random.rand(len(states)) * eps_max)

    greedy = Greedy(env)
    greedy.add_instances(states, goals, eps_l=eps)
    for _ in range(num_steps):
        greedy.step(heuristic_fn, times=times, rand_seen=True)

    trajs: List[List[Tuple[State, float]]] = [instance.traj for instance in greedy.instances]

    trajs_flat: List[Tuple[State, float]]
    trajs_flat, _ = misc_utils.flatten(trajs)

    states_update: List[State] = []
    cost_to_go_update_l: List[float] = []
    for traj in trajs_flat:
        states_update.append(traj[0])
        cost_to_go_update_l.append(traj[1])

    goals_update: List[Goal] = []
    for goal, traj_i in zip(goals, trajs):
        goals_update.extend([goal] * len(traj_i))

    cost_to_go_update = np.array(cost_to_go_update_l)

    is_solved: NDArray[np.bool_] = np.array([instance.is_solved for instance in greedy.instances])

    return states_update, goals_update, cost_to_go_update, is_solved


def update_runner(num_states: int, step_max: int, step_probs: List[float], update_batch_size: int, heur_fn_q: HeurFnQ,
                  env: Environment, result_queue, solve_steps: int, update_method: str, eps_max: float):
    heuristic_fn = heur_fn_q.get_heuristic_fn(env)
    states_per_inst: float = solve_steps
    num_states_curr: int = 0

    states_per_inst_l: List[float] = []
    while num_states_curr < num_states:
        times: Times = Times()
        num_states_i_pred: int = int(np.ceil(update_batch_size * states_per_inst))
        if (num_states_curr + num_states_i_pred) > num_states:
            num_inst_i = int(np.ceil((num_states - num_states_curr) / states_per_inst))
        else:
            num_inst_i = update_batch_size
        num_steps_l: NDArray[np.int_] = np.random.choice(step_max + 1, size=num_inst_i, p=step_probs)

        times_states: Times = Times()
        states, states_goal_set = env.get_start_goal_pairs(list(num_steps_l), times=times_states)
        times.add_times(times_states, ["get_states"])

        if update_method.upper() == "GREEDY":
            times_greedy: Times = Times()
            states_update, goals_update, cost_to_go_update, is_solved = greedy_update(states, states_goal_set, env,
                                                                                      solve_steps, heuristic_fn,
                                                                                      eps_max, times_greedy)
            times.add_times(times_greedy, ["greedy"])
        else:
            raise ValueError("Unknown update method %s" % update_method)

        states_per_inst_l.append(len(states_update) / len(states))
        states_per_inst = float(np.mean(states_per_inst_l))

        leftover: int = (len(states_update) + num_states_curr) - num_states
        if leftover > 0:
            num_keep: int = len(states_update) - leftover
            keep_idxs = np.random.choice(len(states_update), size=num_keep)

            states_update = [states_update[keep_idx] for keep_idx in keep_idxs]
            goals_update = [goals_update[keep_idx] for keep_idx in keep_idxs]
            cost_to_go_update = cost_to_go_update[keep_idxs]

        start_time = time.time()
        states_goals_update_nnet = env.states_goals_to_nnet_input(states_update, goals_update)
        times.record_time("states_goals_to_nnet", time.time() - start_time)

        result_queue.put((states_goals_update_nnet, cost_to_go_update, is_solved, times))

        num_states_curr += len(states_update)

    result_queue.put((None, heur_fn_q.proc_id))


class Updater:
    def __init__(self, env: Environment, num_states: int, back_max: int, step_probs: List[float],
                 heur_fn_qs: List[HeurFnQ], solve_steps: int, update_method: str, update_batch_size: int = 1000,
                 eps_max: float = 0.0):
        super().__init__()
        ctx = get_context("spawn")
        num_procs = len(heur_fn_qs)

        # initialize queues
        self.result_queue: Queue = ctx.Queue()

        # num states per process
        self.num_states: int = num_states
        num_states_per_proc: List[int] = misc_utils.split_evenly(num_states, num_procs)

        # initialize processes
        self.procs: List[BaseProcess] = []
        for proc_id, heur_fn_q in enumerate(heur_fn_qs):
            num_states_proc: int = num_states_per_proc[proc_id]
            if num_states_proc == 0:
                continue
            # update_runner(num_states_proc, back_max, step_probs, update_batch_size,
            #              heur_fn_q, env, self.result_queue, solve_steps, update_method, eps_max)

            proc = ctx.Process(target=update_runner, args=(num_states_proc, back_max, step_probs, update_batch_size,
                                                           heur_fn_q, env, self.result_queue, solve_steps,
                                                           update_method, eps_max))
            proc.daemon = True
            proc.start()
            self.procs.append(proc)

    def update(self) -> Tuple[List[NDArray[Any]], NDArray[np.float64], NDArray[np.bool_]]:
        states_goals_update_nnet: List[NDArray[Any]]
        cost_to_go_update: NDArray[np.float64]
        is_solved: NDArray[np.bool_]
        states_goals_update_nnet, cost_to_go_update, is_solved = self._update()

        output_update = np.expand_dims(cost_to_go_update, 1)

        return states_goals_update_nnet, output_update, is_solved

    def _update(self) -> Tuple[List[NDArray[Any]], NDArray[np.float64], NDArray[np.bool_]]:
        # process results
        states_goals_update_nnet_l: List[List[NDArray[Any]]] = []
        cost_to_go_update_l: List[NDArray[np.float64]] = []
        is_solved_l: List[NDArray[np.bool_]] = []

        none_count: int = 0
        result_count: int = 0
        display_counts: NDArray[np.int_] = np.linspace(1, self.num_states, 10, dtype=int)

        start_time = time.time()
        times: Times = Times()

        num_states_curr: int = 0
        not_done = set(range(len(self.procs)))
        while none_count < len(self.procs):
            # print(num_states, result_count, self.num_batches, none_count, len(self.procs), not_done)
            result = self.result_queue.get()
            if result[0] is None:
                none_count += 1
                not_done.remove(result[1])
                continue
            result_count += 1

            states_goals_nnet_q: List[NDArray[Any]]
            states_goals_nnet_q, cost_to_go_q, is_solved_q, times_q = result

            states_goals_update_nnet_l.append(states_goals_nnet_q)

            cost_to_go_update_l.append(cost_to_go_q)
            is_solved_l.append(is_solved_q)
            num_states_curr += cost_to_go_q.shape[0]
            times.add_times(times_q)

            if num_states_curr >= min(display_counts):
                print(f"Generated {num_states_curr}/{self.num_states} states "
                      f"(%.2f%%) (Total time: %.2f)" % (100 * num_states_curr / self.num_states,
                                                        time.time() - start_time))
                print(times.get_time_str())
                display_counts = display_counts[num_states_curr < display_counts]

        states_goals_update_nnet: List[NDArray[Any]] = []
        for np_idx in range(len(states_goals_update_nnet_l[0])):
            states_goals_nnet_idx: NDArray[Any] = np.concatenate([x[np_idx] for x in states_goals_update_nnet_l],
                                                                 axis=0)
            states_goals_update_nnet.append(states_goals_nnet_idx)

        cost_to_go_update: NDArray[np.float64] = np.concatenate(cost_to_go_update_l, axis=0)
        is_solved: NDArray[np.bool_] = np.concatenate(is_solved_l, axis=0)

        for proc in self.procs:
            proc.join()

        return states_goals_update_nnet, cost_to_go_update, is_solved



from deepxube.environments.environment_abstract import Environment
import math
import re


def get_environment(env_name: str) -> Environment:
    env_name = env_name.lower()
    puzzle_n_regex = re.search(r"puzzle(\d+)", env_name)

    if re.match('^cube3$', env_name):
        from deepxube.environments.cube3 import Cube3
        return Cube3(env_name)
    elif puzzle_n_regex is not None:
        from deepxube.environments.n_puzzle import NPuzzle
        puzzle_dim: int = int(math.sqrt(int(puzzle_n_regex.group(1)) + 1))
        return NPuzzle(env_name, puzzle_dim)
    elif env_name == 'sokoban':
        from deepxube.environments.sokoban import Sokoban
        return Sokoban(env_name)
    else:
        raise ValueError('No known environment %s' % env_name)



from typing import Any, List, Optional, Union
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from torch import Tensor
import numpy as np


class SPLASH(nn.Module):
    def __init__(self, num_hinges: int = 5, init: str = "RELU"):
        super().__init__()
        assert num_hinges > 0, "Number of hinges should be greater than zero, but is %s" % num_hinges
        assert ((num_hinges + 1) % 2) == 0, "Number of hinges should be odd, but is %s" % num_hinges
        init = init.upper()

        self.num_hinges: int = num_hinges
        self.num_each_side: int = int((self.num_hinges + 1) / 2)

        self.hinges: List[float] = list(np.linspace(0, 2.5, self.num_each_side))

        self.output_bias: Parameter = Parameter(torch.zeros(1), requires_grad=True)

        self.coeffs_right: Parameter
        self.coeffs_left: Parameter
        if init == "RELU":
            self.coeffs_right = Parameter(torch.cat((torch.ones(1), torch.zeros(self.num_each_side - 1))),
                                          requires_grad=True)
            self.coeffs_left = Parameter(torch.zeros(self.num_each_side), requires_grad=True)
        elif init == "LINEAR":
            self.coeffs_right = Parameter(torch.cat((torch.ones(1), torch.zeros(self.num_each_side - 1))),
                                          requires_grad=True)
            self.coeffs_left = Parameter(torch.cat((-torch.ones(1), torch.zeros(self.num_each_side - 1))),
                                         requires_grad=True)
        else:
            raise ValueError("Unknown init %s" % init)

    def forward(self, x):
        output: Tensor = torch.zeros_like(x)

        # output for x > 0
        for idx in range(self.num_each_side):
            output = output + self.coeffs_right[idx] * torch.clamp(x - self.hinges[idx], min=0)

        # output for x < 0
        for idx in range(self.num_each_side):
            output = output + self.coeffs_left[idx] * torch.clamp(-x - self.hinges[idx], min=0)

        output = output + self.output_bias

        return output


class LinearAct(nn.Module):
    def __init__(self):
        super().__init__()
        self.dummy = 1

    def forward(self, x):
        self.dummy = 1  # so PyCharm does not complain
        return x


def get_act_fn(act: str):
    act = act.upper()
    if act == "RELU":
        return nn.ReLU()
    elif act == "ELU":
        return nn.ELU()
    elif act == "SIGMOID":
        return nn.Sigmoid()
    elif act == "TANH":
        return nn.Tanh()
    elif act == "SPLASH":
        return SPLASH()
    elif act == "LINEAR":
        return LinearAct()
    else:
        raise ValueError("Un-defined activation type %s" % act)


class ResnetModel(nn.Module):
    def __init__(self, resnet_dim: int, num_resnet_blocks: int, out_dim: int, batch_norm: bool,
                 weight_norm: bool = False, layer_act: str = "RELU"):
        super().__init__()
        self.blocks = nn.ModuleList()
        self.act_fns = nn.ModuleList()

        # resnet blocks
        for block_num in range(num_resnet_blocks):
            block_net = FullyConnectedModel(resnet_dim, [resnet_dim] * 2, [batch_norm] * 2, [layer_act, "LINEAR"],
                                            weight_norms=[weight_norm] * 2)
            module_list: nn.ModuleList = nn.ModuleList([block_net])

            self.blocks.append(module_list)
            self.act_fns.append(get_act_fn(layer_act))

        # output
        self.fc_out = nn.Linear(resnet_dim, out_dim)

    def forward(self, x):
        # resnet blocks
        module_list: nn.ModuleList
        for module_list in self.blocks:
            res_inp = x
            for module in module_list:
                x = module(x)

            x = F.relu(x + res_inp)

        # output
        x = self.fc_out(x)
        return x


class FullyConnectedModel(nn.Module):
    def _forward_unimplemented(self, *input_val: Any) -> None:
        pass

    def __init__(self, input_dim: int, layer_dims: List[int], layer_batch_norms: List[bool], layer_acts: List[str],
                 weight_norms: Optional[List[bool]] = None):
        super().__init__()
        if weight_norms is None:
            weight_norms = [False] * len(layer_dims)
        self.layers: nn.ModuleList = nn.ModuleList()

        # layers
        for layer_dim, batch_norm, act, weight_norm in zip(layer_dims, layer_batch_norms, layer_acts, weight_norms):
            module_list = nn.ModuleList()

            # linear
            if weight_norm:
                module_list.append(nn.utils.weight_norm(nn.Linear(input_dim, layer_dim)))
            else:
                module_list.append(nn.Linear(input_dim, layer_dim))

            # batch norm
            if batch_norm:
                module_list.append(nn.BatchNorm1d(layer_dim))

            # activation
            module_list.append(get_act_fn(act))
            self.layers.append(module_list)

            input_dim = layer_dim

    def forward(self, x):
        x = x.float()

        module_list: nn.ModuleList
        for module_list in self.layers:
            for module in module_list:
                x = module(x)

        return x


class Conv2dModel(nn.Module):
    def _forward_unimplemented(self, *input_val: Any) -> None:
        pass

    def __init__(self, chan_in: int, channel_sizes: List[int], kernel_sizes: List[int], paddings: List[int],
                 layer_batch_norms: List[bool], layer_acts: List[str], strides: Optional[List[int]] = None,
                 transpose: bool = False, weight_norms: Optional[List[bool]] = None,
                 dropouts: Optional[List[float]] = None):
        super().__init__()
        self.layers: nn.ModuleList = nn.ModuleList()
        if strides is None:
            strides = [1] * len(channel_sizes)

        if weight_norms is None:
            weight_norms = [False] * len(channel_sizes)

        if dropouts is None:
            dropouts = [0.0] * len(channel_sizes)

        # layers
        for chan_out, kernel_size, padding, batch_norm, act, stride, weight_norm, dropout in \
                zip(channel_sizes, kernel_sizes, paddings, layer_batch_norms, layer_acts, strides, weight_norms,
                    dropouts):

            module_list = nn.ModuleList()

            # linear
            conv_layer: Union[nn.Conv2d, nn.ConvTranspose2d]
            if transpose:
                conv_layer = nn.ConvTranspose2d(chan_in, chan_out, kernel_size, padding=padding, stride=stride)
            else:
                conv_layer = nn.Conv2d(chan_in, chan_out, kernel_size, padding=padding, stride=stride)

            if weight_norm:
                conv_layer = nn.utils.weight_norm(conv_layer)

            module_list.append(conv_layer)

            # batch norm
            if batch_norm:
                module_list.append(nn.BatchNorm2d(chan_out))

            # activation
            module_list.append(get_act_fn(act))

            # dropout
            if dropout > 0.0:
                module_list.append(nn.Dropout(dropout))

            self.layers.append(module_list)

            chan_in = chan_out

    def forward(self, x):
        x = x.float()

        module_list: nn.ModuleList
        for module_list in self.layers:
            for module in module_list:
                x = module(x)

        return x



from typing import List, Tuple, Optional, Set, Dict, FrozenSet


class Literal:
    def __init__(self, predicate: str, arguments: Tuple[str, ...], directions: Tuple[str, ...], positive: bool = True):
        self.predicate: str = predicate
        self.arguments: Tuple[str, ...] = arguments
        self.directions: Tuple[str, ...] = directions
        for direction in self.directions:
            assert direction in ["in", "out"], f"Direction must be 'in' or 'out' but is {direction}"

        self.arity: int = len(self.arguments)
        self.positive: bool = positive

        self.inputs: Set[str] = set(arg for direction, arg in zip(self.directions, self.arguments) if direction == 'in')
        self.outputs: Set[str] = set(arg for direction, arg in zip(self.directions, self.arguments)
                                     if direction == 'out')

        self.in_out: Set[Tuple[str, str]] = set()
        for arg, direction in zip(self.arguments, self.directions):
            self.in_out.add((arg, direction))

    def to_code(self) -> str:
        prefix: str = ""
        if not self.positive:
            prefix = "not "

        if len(self.arguments) > 0:
            return f'{prefix}{self.predicate}({",".join(self.arguments)})'
        else:
            return f'{prefix}{self.predicate}'

    def get_pred_arity_pos_id(self) -> Tuple[str, int, bool]:
        tup: Tuple[str, int, bool] = (self.predicate, len(self.arguments), self.positive)
        return tup

    def __str__(self):
        return self.to_code()

    def __repr__(self):
        return self.__str__()


class VarNode:
    def __init__(self):
        self.rep: int = 0
        self.neighbors: List[VarNode] = []

    def add_neighbor(self, neighbor):
        self.neighbors.append(neighbor)


class LitNode:
    def __init__(self, predicate: str, in_body: bool, arguments: Tuple[str, ...]):
        self.rep: int = 0
        self.predicate: str = predicate
        self.in_body: bool = in_body
        self.var_nodes: List[VarNode] = []
        self.var_names: Tuple[str, ...] = arguments
        for _ in range(len(arguments)):
            self.var_nodes.append(VarNode())

    def prop_up(self):
        self.rep = hash((self.predicate, self.in_body) + tuple(x.rep for x in self.var_nodes))

    def prop_down(self):
        for var_idx, var_node in enumerate(self.var_nodes):
            var_node.rep = hash((self.rep, var_idx))


def prop_across(var_nodes: List[VarNode]):
    reps_new: List[int] = []
    for var_node in var_nodes:
        rep_new: int = sum([x.rep for x in var_node.neighbors])
        reps_new.append(rep_new)

    for var_node, rep_new in zip(var_nodes, reps_new):
        var_node.rep = rep_new


class Clause:
    def __init__(self, head: Literal, body: Tuple[Literal, ...]):
        self.head: Literal = head
        self.body: Tuple[Literal, ...] = body

        # compute all the 'vars' in the program
        self.all_vars: Set[str] = set()
        self.all_vars.update(head.arguments)
        for literal in (self.head,) + self.body:
            self.all_vars.update(literal.arguments)

        self.hash: Optional[int] = None

    def is_in_out_consistent_body(self) -> bool:
        var_has_out: Set[str] = set()

        for body_lit in self.body:
            var_has_out.update(body_lit.outputs)

        for body_lit in self.body:
            for in_var in body_lit.inputs:
                if in_var not in var_has_out:
                    return False

        return True

    def can_ground(self) -> bool:
        grounded_variables = self.head.inputs
        body_literals = set(self.body)

        while len(body_literals) > 0:
            selected_literal = None
            for literal in body_literals:
                if literal.inputs.issubset(grounded_variables):
                    selected_literal = literal
                    break

            if selected_literal is None:
                return False

            grounded_variables = grounded_variables.union(selected_literal.outputs)
            body_literals = body_literals.difference({selected_literal})

        return True

    def to_code(self) -> str:
        if len(self.body) > 0:
            return (
                f'{self.head.to_code()}:- '
                f'{",".join([blit.to_code() for blit in self.body])}'
            )
        else:
            return self.head.to_code()

    def get_lit_id_count_dict(self) -> Dict[Tuple[str, int, bool], int]:
        lit_pred_dict: Dict[Tuple[str, int, bool], int] = dict()
        for lit in [self.head] + list(self.body):
            if lit is not None:
                lit_tup: Tuple[str, int, bool] = lit.get_pred_arity_pos_id()
                if lit_tup not in lit_pred_dict:
                    lit_pred_dict[lit_tup] = 0

                lit_pred_dict[lit_tup] += 1

        return lit_pred_dict

    def theta_sub(self, other, subs_prev: Optional[Dict[str, str]] = None, negate_l: Optional[List[bool]] = None,
                  subs_forbid: Optional[Dict[str, List[str]]] = None,
                  ignore_head: bool = False) -> Optional[Dict[str, str]]:
        # Initialize
        if ignore_head:
            assert negate_l is None, "Negate not yet integrated with ignore_head"

        if subs_prev is None:
            subs_prev = dict()
        if subs_forbid is None:
            subs_forbid = dict()

        lits_self: List[Literal] = list(self.body)
        lits_other: List[Literal] = list(other.body)
        if not ignore_head:
            lits_self = [self.head] + lits_self
            lits_other = [other.head] + lits_other

        name_to_lit_other: Dict[str, List[Literal]] = dict()
        for lit in lits_other:
            if lit.predicate not in name_to_lit_other.keys():
                name_to_lit_other[lit.predicate] = []
            name_to_lit_other[lit.predicate].append(lit)

        negate_prev = negate_l
        if negate_l is None:
            negate_l = [False] * len(lits_self)

        # check all predicate names in self appears in other
        assert len(lits_self) == len(negate_l), f"{self}, {other}, {len(lits_self)}, {len(negate_l)}, {negate_prev}"
        for lit_self, negate in zip(lits_self, negate_l):
            if (not negate) and (lit_self.predicate not in name_to_lit_other.keys()):
                return None

        return theta_sub_lits(lits_self, name_to_lit_other, negate_l, subs_prev, subs_forbid)

    def __str__(self):
        if len(self.body) > 0:
            # return (
            #    f'{self.head.to_code()}:- '
            #    f'{",".join([blit.to_code() for blit in self.body if blit.predicate[:4] != "dif_"])}'
            # )
            return (
                f'{self.head.to_code()}:- '
                f'{",".join([blit.to_code() for blit in self.body])}'
            )
        else:
            return self.head.to_code()

        # return self.to_code()

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        if self.hash is not None:
            return self.hash

        # make lit nodes
        lit_nodes: List[LitNode] = []
        for lit in [self.head]:
            lit_nodes.append(LitNode(lit.predicate, False, lit.arguments))

        for lit in self.body:
            lit_nodes.append(LitNode(lit.predicate, True, lit.arguments))

        # make connections between variable nodes
        for lit1_idx, lit_node1 in enumerate(lit_nodes):
            for lit2_idx, lit_node2 in enumerate(lit_nodes):
                if lit1_idx == lit2_idx:
                    continue
                for var_idx1, var_name1 in enumerate(lit_node1.var_names):
                    for var_idx2, var_name2 in enumerate(lit_node2.var_names):
                        if var_name1 == var_name2:
                            lit_node1.var_nodes[var_idx1].add_neighbor(lit_node2.var_nodes[var_idx2])

        # get all variable nodes
        var_nodes: List[VarNode] = []
        for lit_node in lit_nodes:
            var_nodes.extend(lit_node.var_nodes)

        # init representation
        for lit_node in lit_nodes:
            lit_node.prop_up()

        # propagate
        self.hash = 0
        for _ in range(10):
            for lit_node in lit_nodes:
                lit_node.prop_down()

            prop_across(var_nodes)

            for lit_node in lit_nodes:
                lit_node.prop_up()

        self.hash = sum([lit_node.rep for lit_node in lit_nodes])

        return self.hash

    def __eq__(self, other):
        # assuming no repeats
        if (self.head is None) != (other.head is None):
            return False
        if self.head.predicate != other.head.predicate:
            return False

        if self.theta_sub(other) is None:
            return False

        if other.theta_sub(self) is None:
            return False

        return True


def prune_lit(lit: Literal, lit_to_prune: Literal, idxs_vars_req: List[Tuple[int, str]]) -> bool:
    if lit.positive != lit_to_prune.positive:
        return True
    else:
        for idx, var_req in idxs_vars_req:
            if lit_to_prune.arguments[idx] != var_req:
                return True
    return False


def theta_sub_lits(lits1: List[Literal], lits2_dict: Dict[str, List[Literal]], negate_l: List[bool],
                   subs_prev: Dict[str, str], subs_forbid: Dict[str, List[str]]) -> Optional[Dict[str, str]]:
    # TODO handle negative literals
    if len(lits1) == 0:
        return subs_prev

    # TODO, negation as failure not really theta subsumption
    lit1: Literal = lits1[0]
    negate: bool = negate_l[0]

    lits2: List[Literal] = lits2_dict.get(lit1.predicate, [])
    if len(lits2) == 0:
        if negate:
            return theta_sub_lits(lits1[1:], lits2_dict, negate_l[1:], subs_prev, subs_forbid)
        else:
            return None

    idxs_not_subbed: List[int] = []
    idxs_vars_req: List[Tuple[int, str]] = []
    for i, x in enumerate(lit1.arguments):
        var_sub: Optional[str] = subs_prev.get(x)
        if var_sub is not None:
            idxs_vars_req.append((i, var_sub))
        else:
            idxs_not_subbed.append(i)

    lits2 = [x for x in lits2 if not prune_lit(lit1, x, idxs_vars_req)]
    if len(lits2) == 0:
        if negate:
            return theta_sub_lits(lits1[1:], lits2_dict, negate_l[1:], subs_prev, subs_forbid)
        else:
            return None

    subs_rec: Optional[Dict[str, str]]
    if len(idxs_not_subbed) == 0:
        # succeeded somewhere and no substitutions needed
        if negate:
            return None
        else:
            subs_rec = theta_sub_lits(lits1[1:], lits2_dict, negate_l[1:], subs_prev, subs_forbid)
            if subs_rec is not None:
                return subs_rec
    else:
        for lit2 in lits2:
            subs: Optional[Dict[str, str]] = theta_sub_args(lit1.arguments, lit2.arguments, idxs_not_subbed,
                                                            subs_prev, subs_forbid)
            if subs is not None:
                if negate:
                    return None
                else:
                    subs_rec = theta_sub_lits(lits1[1:], lits2_dict, negate_l[1:], subs, subs_forbid)
                    if subs_rec is not None:
                        return subs_rec

    if negate:
        return theta_sub_lits(lits1[1:], lits2_dict, negate_l[1:], subs_prev, subs_forbid)
    else:
        return None


def theta_sub_args(args1: Tuple[str, ...], args2: Tuple[str, ...], idxs_not_subbed: List[int],
                   subs_prev: Dict[str, str], subs_forbid: Dict[str, List[str]]) -> Optional[Dict[str, str]]:
    # assuming previous subs checked already if not negate
    subs: Dict[str, str] = subs_prev.copy()

    for idx in idxs_not_subbed:
        arg1: str = args1[idx]
        arg2: str = args2[idx]

        arg1_sub: Optional[str] = subs.get(arg1)
        if (arg1_sub is not None) and (arg1_sub != arg2):
            return None
        else:
            # check if neq constraint not violated
            args_other: Optional[List[str]] = subs_forbid.get(arg1, [])
            if args_other is not None:
                for arg_other in args_other:
                    arg_other_sub: Optional[str] = subs.get(arg_other)
                    if (arg_other_sub is not None) and (arg_other_sub == arg2):
                        return None

            is_var: bool = arg1[0].isupper() and arg1[0].isalpha()
            if is_var:
                # is variable, make substitution
                subs[arg1] = arg2
            else:
                # not variable, cannot substitute
                if arg1 != arg2:
                    return None

    return subs


def make_subs_lit(lit: Literal, subs: Dict[str, str]) -> Literal:
    args_sub: Tuple[str, ...] = tuple(subs[x] for x in lit.arguments)
    lit_sub: Literal = Literal(lit.predicate, args_sub, lit.directions, positive=lit.positive)
    return lit_sub


def make_subs(clause: Clause, subs: Dict[str, str]) -> Clause:
    head_sub: Literal = make_subs_lit(clause.head, subs)
    body_sub: Tuple[Literal, ...] = tuple(make_subs_lit(x, subs) for x in clause.body)
    clause_sub: Clause = Clause(head_sub, body_sub)
    return clause_sub


def theta_sub_replace(clause1: Clause, clause2: Clause, ignore_head: bool = False) -> Clause:
    subs: Optional[Dict[str, str]] = clause1.theta_sub(clause2, ignore_head=ignore_head)
    if subs is None:
        return Clause(clause2.head, clause2.body)

    clause1 = make_subs(clause1, subs)
    body_new: List[Literal] = [clause1.head]

    # TODO changes in to_code() may break this, what about direction?
    clause1_body_code_set: Set[str] = set(lit.to_code() for lit in clause1.body)
    body_new = body_new + [lit for lit in clause2.body if lit.to_code() not in clause1_body_code_set]

    clause2_new: Clause = Clause(clause2.head, tuple(body_new))
    return clause2_new


# first index is the predicate name and the remaining are the arguments
Atom = Tuple[str, ...]
Model = FrozenSet[Atom]



from typing import List, Tuple, Optional
import numpy as np
from numpy.typing import NDArray
from deepxube.environments.environment_abstract import Environment, State, Action, Goal
from deepxube.utils import misc_utils
from deepxube.utils.timing_utils import Times
import time


def is_valid_soln(state: State, goal: Goal, soln: List[Action], env: Environment) -> bool:
    state_soln: State = state
    for action in soln:
        state_soln = env.next_state([state_soln], [action])[0][0]

    return env.is_solved([state_soln], [goal])[0]


def bellman(states: List[State], goals: List[Goal], heuristic_fn,
            env: Environment,
            times: Optional[Times] = None) -> Tuple[NDArray[np.float64], List[NDArray[np.float64]], List[List[State]],
                                                    List[bool]]:
    if times is None:
        times = Times()

    # expand states
    start_time = time.time()
    states_exp, _, tcs_l = env.expand(states)
    times.record_time("expand", time.time() - start_time)

    # get cost-to-go of expanded states
    start_time = time.time()
    states_exp_flat, split_idxs = misc_utils.flatten(states_exp)

    goals_flat = []
    for goal, state_exp in zip(goals, states_exp):
        goals_flat.extend([goal] * len(state_exp))

    ctg_next_flat: NDArray[np.float64] = heuristic_fn(states_exp_flat, goals_flat)
    times.record_time("heur", time.time() - start_time)

    # is solved
    start_time = time.time()
    is_solved = env.is_solved(states, goals)
    times.record_time("is_solved", time.time() - start_time)

    # backup
    start_time = time.time()
    tcs_flat: NDArray[np.float64] = np.hstack(tcs_l)
    ctg_next_p_tc_flat: NDArray[np.float64] = tcs_flat + ctg_next_flat
    ctg_next_p_tc_l: List[NDArray[np.float64]] = np.split(ctg_next_p_tc_flat, split_idxs)

    ctg_backup: NDArray[np.float64] = np.array([np.min(x) for x in ctg_next_p_tc_l]) * np.logical_not(is_solved)
    times.record_time("backup", time.time() - start_time)

    return ctg_backup, ctg_next_p_tc_l, states_exp, is_solved



from typing import List, Set, Tuple, Dict
from deepxube.logic.logic_objects import Clause, Literal, Atom
from deepxube.utils import misc_utils
import re


def parse_literal(lit_str: str) -> Literal:
    # check for negation
    not_match = re.match(r"\s*not\s+(.*)", lit_str)
    positive: bool
    if not_match is None:
        positive = True
    else:
        positive = False
        lit_str = not_match.group(1).strip()

    # parse literal
    lit_str = misc_utils.remove_all_whitespace(lit_str)
    match = re.match(r"([^(]+)(\((.*)\))?", lit_str)
    assert match is not None
    pred_name: str = match.group(1)
    pred_args: Tuple[str, ...]
    if match.group(3) is not None:
        pred_args = tuple([x.strip() for x in match.group(3).split(",")])
    else:
        pred_args = tuple()

    # TODO, better way to handle directions?
    directions = ["in"] * len(pred_args)

    literal: Literal = Literal(pred_name, pred_args, tuple(directions), positive=positive)
    return literal


def replace_anon_vars(lit: Literal, all_lits: List[Literal]):
    all_args: Set[str] = set()
    for lit2 in [lit] + all_lits:
        all_args.update(lit2.arguments)

    args_new: List[str] = []
    for arg in lit.arguments:
        if arg == "_":
            arg_new: str = "V0"
            arg_new_idx: int = 0
            while arg_new in all_args:
                arg_new_idx += 1
                arg_new = f"V{arg_new_idx}"
            all_args.add(arg_new)

            args_new.append(arg_new)
        else:
            args_new.append(arg)

    lit_ret: Literal = Literal(lit.predicate, tuple(args_new), lit.directions, positive=lit.positive)
    return lit_ret


def parse_clause(constr_str: str) -> Tuple[Clause, Dict[str, List[str]]]:
    head_str, body_str = constr_str.split(":-")
    head_lit = parse_literal(head_str)

    # lits
    body_lit_strs: Tuple[str, ...] = tuple(re.findall(r"[^,]+\([^)]+\)|^\s*[^,]+|[^,)]+\s*$", body_str))
    body_lits: List[Literal] = []
    for body_lit_str in body_lit_strs:
        body_lit = parse_literal(body_lit_str)
        body_lits.append(body_lit)

    # subs forbid
    subs_forbid: Dict[str, List[str]] = dict()
    sub_forbid_strs = re.findall(r"\s*([^,]+)\s*!\s*=\s*([^,]+)\s*", body_str)
    for v1, v2 in sub_forbid_strs:
        v1 = v1.strip()
        v2 = v2.strip()
        if v1 not in subs_forbid:
            subs_forbid[v1] = []
        if v2 not in subs_forbid:
            subs_forbid[v2] = []

        subs_forbid[v1].append(v2)
        subs_forbid[v2].append(v1)

    head_lit = replace_anon_vars(head_lit, [head_lit] + body_lits)
    for idx, body_lit in enumerate(body_lits):
        body_lits[idx] = replace_anon_vars(body_lit, [head_lit] + body_lits)

    clause: Clause = Clause(head_lit, tuple(body_lits))
    return clause, subs_forbid


def copy_clause_with_new_head(clause: Clause, head_name_new: str) -> Clause:
    assert clause.head.positive, "Head should be positive"
    clause_new = Clause(Literal(head_name_new, clause.head.arguments, clause.head.directions,
                                positive=clause.head.positive), clause.body)

    return clause_new


def atom_to_str(atom: Atom) -> str:
    return f"{atom[0]}({','.join(atom[1:])})"



from typing import List, Optional, Set, Tuple, Dict, Callable, Any
from deepxube.logic.logic_objects import Clause, Literal, Atom, Model
from deepxube.logic.logic_utils import copy_clause_with_new_head, atom_to_str

import random
import os
import clingo
from clingo import Control, parse_term, Symbol
import re


def model_to_body(model: Model) -> str:
    return ','.join([atom_to_str(atom) for atom in model])


def on_model_var_vals(m) -> frozenset[str]:
    return frozenset(str(x) for x in m.symbols(shown=True))


def parse_clingo_line(line: str) -> str:
    if re.search(r"\S", line) is None:
        return ""

    line = line.strip()
    if re.search(r"\.$", line) is None:
        line = f"{line}."

    return line


class Spec:
    def __init__(self, goal_true: Optional[List[Clause]] = None, goal_false: Optional[List[Clause]] = None,
                 atoms_true: Optional[List[Atom]] = None, atoms_false: Optional[List[Atom]] = None,
                 models_banned: Optional[List[Model]] = None, num_atoms_gt: Optional[int] = None):
        """
        :param goal_true: Must be true. Clauses must have goal in the head.
        :param goal_false: Must be false. Clauses must have goal in the head.
        :param atoms_true: will only return stable models that are a superset (including equality) of given atoms
        :param atoms_false: will not return stable models that contain given atoms
        :param models_banned: will not return a stable model that is a superset (including equality) of given models
        :param num_atoms_gt: Number of atoms in model found must be greater than given number
        """

        self.goal_true: List[Clause] = []
        self.goal_false: List[Clause] = []
        self.atoms_true: List[Atom] = []
        self.atoms_false: List[Atom] = []
        self.models_banned: List[Model] = []
        if goal_true is not None:
            self.goal_true = goal_true.copy()
        if goal_false is not None:
            self.goal_false = goal_false.copy()
        if atoms_true is not None:
            self.atoms_true = atoms_true.copy()
        if atoms_false is not None:
            self.atoms_false = atoms_false.copy()
        if models_banned is not None:
            self.models_banned = models_banned.copy()

        if num_atoms_gt is not None:
            # add num_atoms_gt to goal
            lit_count_gt: Literal = Literal("count_model_grnd_atoms_gt", (str(num_atoms_gt),), ("in",))
            if len(self.goal_true) > 0:
                goal_true_new = []
                for clause in self.goal_true:
                    clause_new = Clause(clause.head, clause.body + (lit_count_gt,))
                    goal_true_new.append(clause_new)
                self.goal_true = goal_true_new
            else:
                clause_new = Clause(Literal("goal", tuple(), tuple()), (lit_count_gt,))
                self.goal_true = [clause_new]

    def add(self, spec_add: 'Spec') -> 'Spec':
        return Spec(goal_true=self.goal_true + spec_add.goal_true,
                    goal_false=self.goal_false + spec_add.goal_false,
                    atoms_true=self.atoms_true + spec_add.atoms_true,
                    atoms_false=self.atoms_false + spec_add.atoms_false,
                    models_banned=self.models_banned + spec_add.models_banned)


class Solver:
    def __init__(self, ground_atoms: List[Atom], bk: List[str]):
        self.bk: List[str] = bk
        self.goals_added: Dict[frozenset[Clause], str] = dict()
        self.models_banned: List[Model] = []

        self.ground_atoms: List[Atom] = ground_atoms
        grnd_atoms_str_l: List[str] = [atom_to_str(atom) for atom in self.ground_atoms]
        model_grnd_atoms_str: str = f"0 {{ {'; '.join(grnd_atoms_str_l)} }} {len(self.ground_atoms)}"
        grnd_atom_counts: List[str] = []
        for grnd_atom_idx, grnd_atom_str in enumerate(grnd_atoms_str_l):
            grnd_atom_counts.append(f"grnd_atom_count_num({grnd_atom_idx})")
            grnd_atom_counts.append(f"grnd_atom_present({grnd_atom_idx}) :- {grnd_atom_str}")

        count_model_grnd_atoms_str: str = "count_model_grnd_atoms(N) :- N = #count{ V: grnd_atom_present(V) }"
        count_model_grnd_atoms_gt_str: str = ("count_model_grnd_atoms_gt(N) :- grnd_atom_count_num(N), "
                                              "count_model_grnd_atoms(M), M > N")
        minimize_grnd_atoms_str = "#minimize {N: count_model_grnd_atoms(N)}"

        # clingo control
        seed = int.from_bytes(os.urandom(4), 'big')
        arguments_rand = ["--models=1", "--opt-mode=ignore", "--heuristic=Domain", "--dom-mod=5,16", "--rand-prob=1",
                          f"--seed={seed}"]
        arguments_min = ["--models=1", "--opt-mode=optN", "--heuristic=Domain", "--dom-mod=5,16"]
        self.ctl_rand: Control = clingo.Control(arguments=arguments_rand)
        self.ctl_min: Control = clingo.Control(arguments=arguments_min)

        for add_line in bk + grnd_atom_counts + [model_grnd_atoms_str, count_model_grnd_atoms_str,
                                                 count_model_grnd_atoms_gt_str]:
            add_line = parse_clingo_line(add_line)
            if len(add_line) == 0:
                continue
            self.ctl_rand.add('base', [], f"{add_line}\n")
            self.ctl_min.add('base', [], f"{add_line}\n")
        add_line = parse_clingo_line(minimize_grnd_atoms_str)
        self.ctl_min.add('base', [], f"{add_line}\n")

        self.ctl_rand.ground([("base", [])])
        self.ctl_min.ground([("base", [])])

    def get_models(self, spec: Spec, on_model: Callable[[Any], Model], num_models: int, minimal: bool) -> List[Model]:
        """

        :param spec: Specification
        :param on_model: Callable that processes models
        :param num_models: number of models to sample
        :param minimal: if true, only samples minimal models
        :return:
        """
        assumptions: List[Tuple[Symbol, bool]] = self._make_assumptions(spec)

        # get models
        models: List[Model] = []
        for model_itr in range(num_models):
            models_i: List[Model] = []
            self.ctl_rand.solve(assumptions=assumptions, on_model=lambda x: models_i.append(on_model(x)))
            if len(models_i) == 0:
                break
            model_i: Model = random.choice(models_i)

            if minimal:
                model_i = self.sample_minimal_model(spec, model_i, on_model)
                # assert model_i == self.sample_minimal_model_old(spec, model_i)

            models.append(model_i)

            if model_itr < (num_models - 1):
                assumptions_i: List[Tuple[Symbol, bool]] = self._make_assumptions(Spec(models_banned=[model_i]))
                assumptions.extend(assumptions_i)

        return models

    def check_model(self, spec: Spec, model: Model) -> bool:
        """

        :param spec: Specification
        :param model: Model to check
        :return:
        """
        atoms_false: List[Atom] = [atom for atom in self.ground_atoms if atom not in model]
        spec_check: Spec = Spec(atoms_true=list(model), atoms_false=atoms_false)

        assumptions: List[Tuple[Symbol, bool]] = self._make_assumptions(spec)
        assumptions += self._make_assumptions(spec_check)
        models_ret: List[None] = []

        self.ctl_rand.solve(assumptions=assumptions, on_model=lambda x: models_ret.append(None))

        return len(models_ret) > 0

    def sample_minimal_model(self, spec: Spec, model: Model, on_model) -> Model:
        models_min: List[Model] = []
        atoms_false: List[Atom] = [atom for atom in self.ground_atoms if atom not in model]
        spec_min: Spec = Spec(atoms_false=atoms_false)
        assumptions: List[Tuple[Symbol, bool]] = self._make_assumptions(spec.add(spec_min))
        self.ctl_min.solve(assumptions=assumptions, on_model=lambda x: models_min.append(on_model(x)))

        return random.choice(models_min)

    def sample_minimal_model_old(self, spec: Spec, model: Model):
        atoms_l = list(model)
        random.shuffle(atoms_l)
        atoms_true: Set[Atom] = set(atoms_l)
        for atom in atoms_l:
            atoms_true.remove(atom)
            model_new: Model = frozenset(atoms_true)
            if self.check_model(spec, model_new):
                return self.sample_minimal_model_old(spec, model_new)

            atoms_true.add(atom)
        return frozenset(atoms_true)

    def _add_goal(self, goal: List[Clause]) -> str:
        """ Adds all goal clauses with same head

        :param goal:
        :return:
        """
        assert all([x.head.predicate == "goal" for x in goal]), "head should be goal"
        assert all([len(x.head.arguments) == 0 for x in goal]), "head should not have any arguments"

        goal_clauses_set: frozenset[Clause] = frozenset(goal)
        goal_new_head_pred_get: Optional[str] = self.goals_added.get(goal_clauses_set)
        if goal_new_head_pred_get is None:
            goal_num: int = len(self.goals_added)
            goal_new_head_pred: str = f"goal{goal_num}"

            prg_blk: str = goal_new_head_pred
            for goal_clause in goal:
                goal_clause_new_head: Clause = copy_clause_with_new_head(goal_clause, goal_new_head_pred)
                self.ctl_rand.add(prg_blk, [], f"{goal_clause_new_head.to_code()}.\n")
                self.ctl_min.add(prg_blk, [], f"{goal_clause_new_head.to_code()}.\n")
            self.ctl_rand.ground([(prg_blk, [])])
            self.ctl_min.ground([(prg_blk, [])])

            self.goals_added[goal_clauses_set] = goal_new_head_pred

            return goal_new_head_pred
        else:
            return goal_new_head_pred_get

    def _make_assumptions(self, spec: Spec) -> List[Tuple[Symbol, bool]]:
        atoms_true: List[Atom] = []
        atoms_false: List[Atom] = []
        if len(spec.goal_true) > 0:
            atoms_true.append((self._add_goal(spec.goal_true),))
        if len(spec.goal_false) > 0:
            atoms_false.append((self._add_goal(spec.goal_false),))

        atoms_true += spec.atoms_true
        atoms_false += spec.atoms_false
        for model_banned in spec.models_banned:
            blits: List[Literal] = [Literal(atom[0], atom[1:], tuple(["in"] * len(atom[1:]))) for atom in
                                    model_banned]
            clause_banned: Clause = Clause(Literal("goal", tuple(), tuple()), tuple(blits))
            atoms_false.append((self._add_goal([clause_banned]),))

        assumed_true: List[str] = []
        assumed_false: List[str] = []
        for atom in atoms_true:
            assumed_true.append(atom_to_str(atom))
        for atom in atoms_false:
            assumed_false.append(atom_to_str(atom))

        assumptions: List[Tuple[Symbol, bool]] = []
        for lit in assumed_true:
            assumptions.append((parse_term(lit), True))
        for lit in assumed_false:
            assumptions.append((parse_term(lit), False))

        return assumptions



from typing import List, Union

import numpy as np
from matplotlib import pyplot as plt

from deepxube.environments.environment_abstract import Environment, State, Goal


class Quaternion:
    """Quaternion Rotation:

    Class to aid in representing 3D rotations via quaternions.
    """

    @classmethod
    def from_v_theta(cls, v, theta):
        """
        Construct quaternions from unit vectors v and rotation angles theta

        Parameters
        ----------
        v : array_like
            array of vectors, last dimension 3. Vectors will be normalized.
        theta : array_like
            array of rotation angles in radians, shape = v.shape[:-1].

        Returns
        -------
        q : quaternion object
            quaternion representing the rotations
        """
        theta = np.asarray(theta)
        v = np.asarray(v)
        s = np.sin(0.5 * theta)
        c = np.cos(0.5 * theta)

        v = v * s / np.sqrt(np.sum(v * v, -1))
        x_shape = v.shape[:-1] + (4,)

        x = np.ones(x_shape).reshape(-1, 4)
        x[:, 0] = c.ravel()
        x[:, 1:] = v.reshape(-1, 3)
        x = x.reshape(x_shape)

        return cls(x)

    def __init__(self, x):
        self.x = np.asarray(x, dtype=float)

    def __repr__(self):
        return "Quaternion:\n" + self.x.__repr__()

    def __mul__(self, other):
        # multiplication of two quaternions.
        # we don't implement multiplication by a scalar
        sxr = self.x.reshape(self.x.shape[:-1] + (4, 1))
        oxr = other.x.reshape(other.x.shape[:-1] + (1, 4))

        prod = sxr * oxr
        return_shape = prod.shape[:-1]
        prod = prod.reshape((-1, 4, 4)).transpose((1, 2, 0))

        ret = np.array([(prod[0, 0] - prod[1, 1]
                         - prod[2, 2] - prod[3, 3]),
                        (prod[0, 1] + prod[1, 0]
                         + prod[2, 3] - prod[3, 2]),
                        (prod[0, 2] - prod[1, 3]
                         + prod[2, 0] + prod[3, 1]),
                        (prod[0, 3] + prod[1, 2]
                         - prod[2, 1] + prod[3, 0])],
                       dtype=float,
                       order='F').T
        return self.__class__(ret.reshape(return_shape))

    def as_v_theta(self):
        """Return the v, theta equivalent of the (normalized) quaternion"""
        x = self.x.reshape((-1, 4)).T

        # compute theta
        norm = np.sqrt((x ** 2).sum(0))
        theta = 2 * np.arccos(x[0] / norm)

        # compute the unit vector
        v = np.array(x[1:], order='F', copy=True)
        v /= np.sqrt(np.sum(v ** 2, 0))

        # reshape the results
        v = v.T.reshape(self.x.shape[:-1] + (3,))
        theta = theta.reshape(self.x.shape[:-1])

        return v, theta

    def as_rotation_matrix(self):
        """Return the rotation matrix of the (normalized) quaternion"""
        v, theta = self.as_v_theta()

        shape = theta.shape
        theta = theta.reshape(-1)
        v = v.reshape(-1, 3).T
        c = np.cos(theta)
        s = np.sin(theta)

        mat = np.array([[v[0] * v[0] * (1. - c) + c,
                         v[0] * v[1] * (1. - c) - v[2] * s,
                         v[0] * v[2] * (1. - c) + v[1] * s],
                        [v[1] * v[0] * (1. - c) + v[2] * s,
                         v[1] * v[1] * (1. - c) + c,
                         v[1] * v[2] * (1. - c) - v[0] * s],
                        [v[2] * v[0] * (1. - c) - v[1] * s,
                         v[2] * v[1] * (1. - c) + v[0] * s,
                         v[2] * v[2] * (1. - c) + c]],
                       order='F').T
        return mat.reshape(shape + (3, 3))

    def rotate(self, points):
        rot_mat = self.as_rotation_matrix()
        return np.dot(points, rot_mat.T)


def project_points(points, q, view, vertical):  # type: ignore
    """Project points using a quaternion q and a view v

    Parameters
    ----------
    points : array_like
        array of last-dimension 3
    q : utils.viz_utils.Quaternion
        quaternion representation of the rotation
    view : array_like
        length-3 vector giving the point of view
    vertical : array_like
        direction of y-axis for view.  An error will be raised if it
        is parallel to the view.

    Returns
    -------
    proj: array_like
        array of projected points: same shape as points.
    """
    if vertical is None:
        vertical = [0, 1, 0]
    points = np.asarray(points)   # type: ignore
    view = np.asarray(view)

    xdir = np.cross(vertical, view).astype(float)   # type: ignore

    if np.all(xdir == 0):
        raise ValueError("vertical is parallel to v")

    xdir /= np.sqrt(np.dot(xdir, xdir))

    # get the unit vector corresponing to vertical
    ydir = np.cross(view, xdir)   # type: ignore
    ydir /= np.sqrt(np.dot(ydir, ydir))

    # normalize the viewer location: this is the z-axis
    v2 = np.dot(view, view)
    zdir = view / np.sqrt(v2)

    # rotate the points
    rot_mat = q.as_rotation_matrix()
    r_pts = np.dot(points, rot_mat.T)

    # project the points onto the view
    dpoint = r_pts - view
    dpoint_view = np.dot(dpoint, view).reshape(dpoint.shape[:-1] + (1,))
    dproj = -dpoint * v2 / dpoint_view

    trans = list(range(1, dproj.ndim)) + [0]
    return np.array([np.dot(dproj, xdir),
                     np.dot(dproj, ydir),
                     -np.dot(dpoint, zdir)]).transpose(trans)


def visualize_examples(env: Environment, states: Union[List[State], List[Goal]]):
    states_np = env.visualize(states)

    plt.ion()
    fig = plt.figure()

    ax1 = fig.add_subplot(221)
    ax2 = fig.add_subplot(222)
    ax3 = fig.add_subplot(223)
    ax4 = fig.add_subplot(224)

    idx_stop: int
    if states_np.shape[3] == 3:
        axs = [ax1, ax2, ax3, ax4]

        for idx in range(0, len(states_np), 4):
            for ax in axs:
                ax.cla()

            idx_stop = min(idx + 4, len(states_np))
            for idx_ax, idx_show in enumerate(range(idx, idx_stop)):
                ax = axs[idx_ax]
                ax.imshow(states_np[idx_show])
                ax.set_xticks([])
                ax.set_yticks([])

            fig.canvas.draw()
            input("Enter anything: ")

    elif states_np.shape[3] == 6:
        axs1 = [ax1, ax2]
        axs2 = [ax3, ax4]
        axs_cube = [axs1, axs2]

        for idx in range(0, len(states_np), 2):
            for axs in axs_cube:
                for ax in axs:
                    ax.cla()

            idx_stop = min(idx + 2, len(states_np))
            for idx_ax, idx_show in enumerate(range(idx, idx_stop)):
                axs = axs_cube[idx_ax]
                axs[0].imshow(states_np[idx_show, :, :, :3])
                axs[1].imshow(states_np[idx_show, :, :, 3:])

                for ax in axs:
                    ax.set_xticks([])
                    ax.set_yticks([])

            fig.canvas.draw()
            input("Enter anything: ")

    plt.ioff()
    plt.close(fig)



from typing import List, Any

import torch
import torch.nn as nn
from torch.multiprocessing import Queue, get_context

from deepxube.environments.environment_abstract import Environment, State
from deepxube.nnet import nnet_utils
from deepxube.utils.misc_utils import flatten
import numpy as np
from numpy.typing import NDArray

import time


def data_runner(queue1: Queue, queue2: Queue):
    while True:
        the = queue1.get()
        if the is None:
            break
        queue2.put(the)


def test_env(env: Environment, num_states: int, step_max: int):
    torch.set_num_threads(1)

    # generate start/goal states
    start_time = time.time()
    states: List[State] = env.get_start_states(num_states)

    elapsed_time = time.time() - start_time
    states_per_sec = len(states) / elapsed_time
    print("Generated %i start states in %s seconds (%.2f/second)" % (len(states), elapsed_time, states_per_sec))

    # get data
    start_time = time.time()
    states, goals = env.get_start_goal_pairs(list(np.random.randint(step_max + 1, size=num_states)))

    elapsed_time = time.time() - start_time
    states_per_sec = len(states) / elapsed_time
    print("Generated %i start/goal states in %s seconds (%.2f/second)" % (len(states), elapsed_time, states_per_sec))

    # is_solved
    start_time = time.time()
    is_solved_l: List[bool] = env.is_solved(states, goals)
    per_solved: float = 100.0 * float(np.mean(is_solved_l))
    elapsed_time = time.time() - start_time
    states_per_sec = len(states) / elapsed_time
    print(f"Computed is_solved for {len(states)} states ({per_solved:.2f}% solved) in {elapsed_time} seconds "
          f"({states_per_sec:.2f}/second)")

    # expand
    start_time = time.time()
    states_exp, _, tcs = env.expand(states)
    ave_next_states: float = float(np.mean([len(x) for x in states_exp]))
    ave_tc: float = float(np.mean(flatten(tcs)[0]))

    elapsed_time = time.time() - start_time
    states_per_sec = len(states) / elapsed_time
    print(f"Expanded %i states, mean #next/tc: ({ave_next_states:.2f}/{ave_tc:.2f}), "
          f"in %s seconds (%.2f/second)" % (len(states), elapsed_time, states_per_sec))

    # nnet format
    start_time = time.time()
    states_goals_nnet: List[NDArray[Any]] = env.states_goals_to_nnet_input(states, goals)
    elapsed_time = time.time() - start_time
    states_per_sec = len(states) / elapsed_time
    print("Converted %i states and goals to nnet format in "
          "%s seconds (%.2f/second)" % (len(states), elapsed_time, states_per_sec))

    # get heuristic fn
    on_gpu: bool
    device: torch.device
    device, devices, on_gpu = nnet_utils.get_device()
    print("device: %s, devices: %s, on_gpu: %s" % (device, devices, on_gpu))

    nnet: nn.Module = env.get_v_nnet()
    nnet.to(device)
    if on_gpu:
        nnet = nn.DataParallel(nnet)

    # initialize nnet
    print("")
    heuristic_fn = nnet_utils.get_heuristic_fn(nnet, device, env)
    heuristic_fn(states, goals)

    # nnet heuristic
    start_time = time.time()
    heuristic_fn(states, goals)

    nnet_time = time.time() - start_time
    states_per_sec = len(states) / nnet_time
    print("Computed heuristic for %i states in %s seconds (%.2f/second)" % (len(states), nnet_time, states_per_sec))

    # multiprocessing
    print("")
    start_time = time.time()
    ctx = get_context("spawn")
    queue1: Queue = ctx.Queue()
    queue2: Queue = ctx.Queue()
    proc = ctx.Process(target=data_runner, args=(queue1, queue2))
    proc.daemon = True
    proc.start()
    print("Process start time: %.2f" % (time.time() - start_time))

    queue1.put((states_goals_nnet,))
    queue2.get()

    start_time = time.time()
    queue1.put((states, goals))
    print("State/goals send time: %s" % (time.time() - start_time))

    start_time = time.time()
    queue2.get()
    print("States/goals get time: %.2f" % (time.time() - start_time))

    start_time = time.time()
    queue1.put((states_goals_nnet,))
    print("State/goals nnet send time: %s" % (time.time() - start_time))

    start_time = time.time()
    queue2.get()
    print("States/goals nnet get time: %.2f" % (time.time() - start_time))

    start_time = time.time()
    queue1.put(env)
    print("Environment send time: %s" % (time.time() - start_time))

    start_time = time.time()
    queue2.get()
    print("Environment get time: %s" % (time.time() - start_time))

    start_time = time.time()
    queue1.put(None)
    proc.join()
    print("Process join time: %.2f" % (time.time() - start_time))



from typing import List, Tuple, Any, Union, Set
import numpy as np
import math
import re
from numpy.typing import NDArray


def flatten(data: List[List[Any]]) -> Tuple[List[Any], List[int]]:
    num_each = [len(x) for x in data]
    split_idxs: List[int] = list(np.cumsum(num_each)[:-1])

    data_flat = [item for sublist in data for item in sublist]

    return data_flat, split_idxs


def unflatten(data: Union[List[Any], NDArray[Any]], split_idxs: List[int]) -> List[List[Any]]:
    data_split: List[List[Any]] = []

    start_idx: int = 0
    end_idx: int
    for end_idx in split_idxs:
        data_split.append(list(data[start_idx:end_idx]))
        start_idx = end_idx

    data_split.append(list(data[start_idx:]))

    return data_split


def split_evenly(num_total: int, num_splits: int) -> List[int]:
    num_per: List[int] = [math.floor(num_total / num_splits) for _ in range(num_splits)]
    left_over: int = num_total % num_splits
    for idx in range(left_over):
        num_per[idx] += 1

    return num_per


def remove_all_whitespace(val: str) -> str:
    pattern = re.compile(r'\s+')
    val = re.sub(pattern, '', val)

    return val


def random_subset(set_orig: Union[Set[Any], frozenset[Any]], keep_prob: bool) -> Set[Any]:
    rand_vals: NDArray[Any] = np.random.rand(len(set_orig))
    keep_arr: NDArray[np.bool_] = np.array(rand_vals < keep_prob)
    rand_subset: Set[Any] = set(elem for elem, keep_i in zip(set_orig, keep_arr) if keep_i)

    return rand_subset


def boltzmann(vals: List[float], temp: float) -> List[float]:
    if len(vals) == 1:
        return [1.0]
    else:
        vals_np: NDArray[np.float64] = np.array(vals)
        exp_vals_np: NDArray[np.float64] = np.exp((1.0 / temp) * (vals_np - np.max(vals_np)))
        probs_np: NDArray[np.float64] = exp_vals_np / np.sum(exp_vals_np)

        return list(probs_np)
